/*
 *
 * O.Filangi : linearisation de la vraissemblance -  calcul simultané des position sur le génome
 * -deviceemu => emulation
 * nvcc -g -arch=sm_20 --cuda analyse_whole-genome.cu
 * debugging : ddd --debugger /usr/local/cuda/bin/cuda-gdb qtlmap
 * Debug   : nvcc -g -G -gencode=arch=compute_20,code=\"sm_20,compute_20\" --cuda analyse_whole-genome.cu
 * Release : nvcc -O4 --use_fast_math  -gencode=arch=compute_20,code=\"sm_20,compute_20\" --cuda analyse_whole-genome.cu
 * 
 * 
 * TODO : XX est une matrice symetrique => utiliser que la moitie superieur de la matrice
 * 
 * pour le profiling : /usr/local/cuda/computeprof/bin/computeprof
 */
#include <stdio.h>
#include <stdlib.h>
#include <cuda.h>
#include <iostream>
using namespace std;

#define safecall(call) do{\
		cudaError_t errERR = call ; \
		if (cudaSuccess != errERR){\
			fprintf(stderr, "cuda error at %s:%d, %s\n",\
					__FILE__, __LINE__, cudaGetErrorString(errERR));exit(1);\
		}\
}while(0) 

#define PRECISION_FLOTTANT 4
/* ------------------------------- ACCESSORS ------------------------------------------ */
#define OPTIMIZE_MEMORY_MATSYM true

#if(!OPTIMIZE_MEMORY_MATSYM)

#define BASE_SCALE_SYMMAT data.ntniv*data.npos*data.ntniv
#define GETACC(data,isim,ipos,iniv,jniv) (iniv>=jniv?isim*BASE_SCALE_SYMMAT+ipos+iniv*data.npos+jniv*data.ntniv*data.npos:isim*BASE_SCALE_SYMMAT+ipos+jniv*data.npos+iniv*data.ntniv*data.npos)
#define SIZEMATSYM(nsim,npos,ntniv) (nsim*npos*ntniv*ntniv)
#else

#define CEIL(VARIABLE) ( (VARIABLE - (int)VARIABLE)==0 ? (int)VARIABLE : (int)VARIABLE+1 )
#define BASE_SCALE_SYMMAT ((data.npos*data.ntniv*(data.ntniv+1)/2))
#define GETACC(data,isim,ipos,iniv,jniv) ( iniv>jniv?ipos+(iniv*(iniv+1)/2+jniv)*data.npos+isim*BASE_SCALE_SYMMAT:ipos+((jniv*(jniv+1))/2+iniv)*data.npos+isim*BASE_SCALE_SYMMAT )
#define SIZEMATSYM(nsim,npos,ntniv) (nsim*npos*( ntniv*(ntniv+1) /2 ))

#endif



#ifdef CUDA_SP
#define DT float
#else
#define DT double
#endif


/* Generate print information about sbu-results */
#if ! defined(_CUDA_HOST_DEBUG_)
#define _CUDA_HOST_DEBUG_ 0
#endif

/* Generate time comsuming information about sub process */
#if ! defined(_CUDA_HOST_TIME_PROF_)
#define _CUDA_HOST_TIME_PROF_ 0
#endif
/* number of thread available by block */
#define MAX_BLOCKDIM_512 512
#define MAX_BLOCKDIM_64 64
#define MAX_BLOCKDIM_32 32
#define BLOCKDIMX_XTX_C 96
#define BLOCKDIMX 128
#define MAXND_WORK 5000
#define MAXNP_WORK 100



#define MAXNTNIV_LOCALTHREAD_MEMORY 200

/* Definition des donnees en memoire constante sur gpu */

__constant__ int constNbLevelFix ;
__constant__ int constNbLevelVar ;
__constant__ int constNSim;
__constant__ int constCorrIndexCol[MAXNTNIV_LOCALTHREAD_MEMORY];
__constant__ int constSizeFam[MAXNP_WORK];


/* Keep in memory the residual variance fitted for each Hypothesis analysis */
static DT * allVarFitted = NULL ;

void print_info_memory() {
	size_t free,total;
	cuMemGetInfo(&free, &total) ;
	printf("===============> free: %f Mb (%lu bytes)   total:%f Mb (%lu bytes) \n",double(free)/(1024*1024),free,double(total)/(1024*1024),total);
}

class  QTLMapStructDeviceDataLinear {
public:
	//DT * contingence ; /* matrice de contingence les effets statiques sont dans les premieres colonnes. */
	DT * contingence_fix ; /*                                      [NDMAX][NFIX]                         */
	DT * contingence_fix_host ;
	DT * contingence_var ; /*                                      [ND][NVAR][NPOS]                   */  
	DT * Y          ; /* Vecteur de Performance Y                  [NSIM][ND]        */
	DT * CD         ; /* Vecteur des CD                            [ND]        */
	int nqtl        ; /* hypothese courante */
	int np          ; /* nombre de famille de pere */
	int nsim        ; /* nombre de simulation */
	int nposGlobal  ; /* nombre de position teste */
	int npos        ; /* nombre de position teste par block */
	int ndmax       ; /* taille du tableau Y */
	int nd          ; /* nombre de descendants a prendre en compte */
	int nLevelFix   ; /* nombre d effet non dependant de la position */
	int nLevelVar   ; /* nombre d'effet dependant de la position */
	int ntniv       ; /* nLevelFix + nLevelVar */
	double seuil_cho    ; /* seuil pour l estimation des effets */
	DT * isigsq     ; /* variance residuelle des Hypothese < NQTL           [NSIM][NP] */
	int *corIpKd    ; /* *correspondance kd -> ip (indice du pere de la progeniture kd) */

public:

	QTLMapStructDeviceDataLinear();

	void init(   
			int * nqtlPtr,
			DT * sigsquare_host,
			DT * xinc_d_fix,
			DT * Y_d,
			DT * CD_d,
			int *corrLevelColPtr,
			double *seuil_choPtr,
			int * ndPtr,
			int * nkdPtr,
			int * nsimPtr,
			int * npositionGlobalPtr,
			//		int * npositionBlockPtr,
			int * nLevelFixPtr,
			int * nLevelVarPtr,
			int * npPtr,
			int * sizeFamilyNp);

	virtual ~QTLMapStructDeviceDataLinear();

	void set_contingence_var(int ndmax, int nLevelVar,int nbPositionsTest, int startPosition, DT* xinc_d_var,cudaStream_t * stream);

	void releaseDevice();

	size_t calculBlockPositionWorkSize(bool hetero) ;
} ;

class QTLMapStructDeviceWorkLinear {
public :
	DT * A_res      ; /* matrice d'incidence                                homosscedastic : [NFIX][NFIX]  , heteroscedastic:  not used ! */

	DT * XX         ; /* matrice d'incidence                                homosscedastic : [NPOS][NTNIV][NTNIV]  , heteroscedastic: [NPOS][NTNIV][NTNIV][NSIM] */
	DT * triang     ; /* matrice temporaire de la decomposition Cholesky    homosscedastic : [NPOS][NTNIV][NTNIV]  , heteroscedastic: [NPOS][NTNIV][NTNIV][NSIM] */
	DT * rhs        ; /* Vecteur RHS                                                         [NPOS][NSIM][NTNIV]  */

public :
	QTLMapStructDeviceWorkLinear();
	/* init de XX, triang, rhs */
	void initResolution(bool hetero,const QTLMapStructDeviceDataLinear & data,DT *CD,DT * sigsquare_host,int * sizeFamily);

	virtual ~QTLMapStructDeviceWorkLinear();

	void releaseDeviceResolution();


	/* recuperation du tableau vecsol stoque sur le device */
	void getAllVecsolDevice(const QTLMapStructDeviceDataLinear & data,int nbBloc,int nbThread,int *vecsol) ;

} ;


/**
 * 
 */
class QTLMapStructDeviceSolutionLinear {
public:
	DT * osigsq     ; /* Vecteur des variances residuelles                  [NPOS][NSIM][NP]     */
	DT * bestim     ; /* Vecteur des solution des effets                    [NPOS][NSIM][NTNIV ] */
	DT * lrt        ; /* rapport de vraisemblance                           [NPOS][NSIM] */
public :

	QTLMapStructDeviceSolutionLinear();

	void init(const QTLMapStructDeviceDataLinear & data);

	virtual ~QTLMapStructDeviceSolutionLinear();

	void releaseDevice();

} ;


class QTLMapGenericModelCalcul {

public: 
	virtual void start_analysis(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) = 0 ;
	virtual bool convergenceOk(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceSolutionLinear solution) = 0 ;
	virtual void calcul_XT_X_A(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) = 0;
	virtual void calcul_XT_X_B(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) = 0;
	virtual void calcul_XT_X_C(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) = 0;
	virtual void calcul_Cholesky_Decomposition(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) = 0;
	virtual void calcul_RHS(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) = 0;
	virtual void calcul_LU(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,QTLMapStructDeviceSolutionLinear solution) = 0;
	virtual void calcul_SIGSQ(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,QTLMapStructDeviceSolutionLinear solution) = 0;
};

class QTLMapHomoscedasticModelCalcul : public QTLMapGenericModelCalcul {

public:
	virtual void start_analysis(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work);
	virtual bool convergenceOk(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceSolutionLinear solution);
	virtual void calcul_XT_X_A(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work);
	virtual void calcul_XT_X_B(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work);
	virtual void calcul_XT_X_C(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work);
	virtual void calcul_Cholesky_Decomposition(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work);
	virtual void calcul_RHS(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work);
	virtual void calcul_LU(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,QTLMapStructDeviceSolutionLinear solution);
	virtual void calcul_SIGSQ(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,QTLMapStructDeviceSolutionLinear solution);
};


class QTLMapHeteroscedasticModelCalcul : public QTLMapGenericModelCalcul {
private :
	DT * lastLrtConvergenceHost ; /* save the last lrt to compute the convergence for each position/simulation */
	DT * lrtConvergenceDevice   ; /* do once an allocation to get lrt from the current analysis */
	DT * varInDevice            ; /* [NPOS][NSIM][NP]*/
public:
	/**
	 * Constructor
	 */
	QTLMapHeteroscedasticModelCalcul(int npos,int nsim,int np) { 
		int size=npos*nsim;
		safecall(cudaMallocHost(&lastLrtConvergenceHost,size*sizeof(DT)));
		safecall(cudaMalloc(&lrtConvergenceDevice,size*sizeof(DT)));

		for (int i=0;i<size;i++)
			lastLrtConvergenceHost[i]=99999.0;

		size=npos*nsim*np;
		safecall(cudaMalloc(&varInDevice,size*sizeof(DT)));
	} ;

	/**
	 * Destructor => delete lastlrt
	 */
	virtual ~QTLMapHeteroscedasticModelCalcul() { 
		safecall(cudaFreeHost(lastLrtConvergenceHost)); 
		safecall(cudaFree(lrtConvergenceDevice)); 
		safecall(cudaFree(varInDevice)); 
	} ;

	virtual void start_analysis(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work);
	virtual bool convergenceOk(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceSolutionLinear solution);
	virtual void calcul_XT_X_A(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work);
	virtual void calcul_XT_X_B(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work);
	virtual void calcul_XT_X_C(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work);
	virtual void calcul_Cholesky_Decomposition(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work);
	virtual void calcul_RHS(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work);
	virtual void calcul_LU(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,QTLMapStructDeviceSolutionLinear solution);
	virtual void calcul_SIGSQ(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,QTLMapStructDeviceSolutionLinear solution);
};

size_t QTLMapStructDeviceDataLinear::calculBlockPositionWorkSize(bool hetero) {

	/* le calcul comprend 
	 * A_Res           : NFIX^2 ou 0 (heteroscedastic)
	 * XX              : NPOS*NTNIV*NTNIV ou NPOS*NTNIV*NTNIV*NSIM (heteroscedastic)
	 * Triang          : NPOS*NTNIV*NTNIV ou NPOS*NTNIV*NTNIV*NSIM (heteroscedastic)
	 * Rhs             : NPOS*NSIM*NTNIV
	 * OSIG            : NPOS*NSIM*NP
	 * BESTIM          : NPOS*NSIM*NTNIV
	 * LRT             : NQTL*NPOSGLOBAL*NSIM
	 * CONTINGENCE_VAR : NPOS*NVAR*NPOS
	 * 
	 * 
	 * Y,CD et contingence_fix ont deja ete alloue 
	 * 
	 */

	size_t free,total;
	cuMemGetInfo(&free, &total) ;
	// avec la version driver 3.7 on etait obliger de diviser par 1.8 pour que ca passe.....;
	free = free / 1.1 ;

	int num  ; 
	if (hetero) {
		num = ( nqtl*nposGlobal*nsim + MAXND_WORK + np*nsim )*sizeof(DT);
	}
	else {
		num = ( nLevelFix*nLevelFix + nqtl*nposGlobal*nsim + MAXND_WORK + np*nsim )*sizeof(DT);
	}
	if ( free < num ) {
		cerr << "Not enough memory free:"<< free << " need:"<< num << endl ;
		exit(1);
	}
	size_t res = free - num;
	size_t denom ;
	if (hetero) {
		denom = (2*SIZEMATSYM(nsim,1,ntniv) + 2*nsim*ntniv+ nsim*np + nd*nLevelVar)*sizeof(DT);
	} else {
		denom = (2*SIZEMATSYM(1,1,ntniv) + 2*nsim*ntniv+ nsim*np + nd*nLevelVar)*sizeof(DT);
	}
	res = ceil(res/denom)+1;

	/* The CUDA methods Set_RHS_VAR have a limited block size in x : 16000 */
	if ( nLevelVar > 0) {
		int sizeCalculSet_Rhs_VAR = (16000 * MAX_BLOCKDIM_64) / nLevelVar ;
		if ( sizeCalculSet_Rhs_VAR < res ) res = sizeCalculSet_Rhs_VAR;
	}

	cout << "**********************************************"<< endl;
	cout << "           **   Block NPOS :"<< res << " **"<< endl;
	cout << "**********************************************"<< endl;
	return res ;

} ;

template <class T>
class Utils {
public:
	static void printFloatDeviceArray1D(int dim1,int nb1,T* array1D);
	static void printFloatDeviceArray2D(int dim1,int dim2,int nb1,int nb2,T* array2D);
	static void printFloatDeviceArray3D(int dim1,int dim2,int dim3,int nb1,int nb2,int nb3,T* array3D);
	static void printFloatDeviceArray4D(int dim1,int dim2,int dim3,int dim4,int nb1,int nb2,int nb3,int nb4,T* array4D);
	static void printFloatHostArray2D(int dim1,int dim2,int nb1,int nb2,T* array2D);
	static void printFloatHostArray3D(int dim1,int dim2,int dim3,int nb1,int nb2,int nb3,T* array3D);
	static void getArrayDeviceToHost(int nbBloc,int nbThread,int size,T * inDeviceArray, T *outHostArray,cudaStream_t * stream);
};

/*******************************************************************************************************************************************/

QTLMapStructDeviceDataLinear::QTLMapStructDeviceDataLinear() {
	contingence_fix = NULL ; 
	contingence_var = NULL ;
	Y = NULL;
	CD = NULL;
	isigsq = NULL;
}


void QTLMapStructDeviceDataLinear::init(
		int * nqtlPtr,
		DT * sigsquare_host,
		DT * xinc_d_fix,
		DT * Y_d,
		DT * CD_d,
		int *corrLevelColPtr,
		double *seuil_choPtr,
		int * ndPtr,
		int * nkdPtr,
		int * nsimPtr,
		int * npositionGlobalPtr,
		//		int * npositionBlockPtr,
		int * nLevelFixPtr,
		int * nLevelVarPtr,
		int * npPtr,
		int * sizeFamilyNp) {

	seuil_cho   = *seuil_choPtr;
	seuil_cho   = sqrt(seuil_cho);
	ndmax       = *ndPtr       ; /* taille reelle de la matrice de contingence : lignes */
	nd          = *nkdPtr      ; /* taille de la population pris en compte */
	nsim        = *nsimPtr     ;
	npos        = 0;
	nposGlobal  = *npositionGlobalPtr;
	nLevelFix   = *nLevelFixPtr ;
	nLevelVar   = *nLevelVarPtr;
	np          = *npPtr;
	ntniv       = nLevelFix + nLevelVar ;
	nqtl        = *nqtlPtr;

#if(_CUDA_HOST_DEBUG_)
	printf("*********************** INIT DATA *********************************\n");
	printf("seuil_cho  SQ : %f\n",seuil_cho);
	printf("ndmax         : %d\n",ndmax);
	printf("nd            : %d\n",nd);
	printf("nsim          : %d\n",nsim);
	printf("np            : %d\n",np);
	printf("nposGlobal    : %d\n",nposGlobal);
	printf("npos          : %d\n",npos);
	printf("nLevelFix     : %d\n",nLevelFix);
	printf("nLevelVar     : %d\n",nLevelVar);
	for (int i=0;i<np;i++) printf("sizeFamilyNp[%d] : %d\n",i,sizeFamilyNp[i]);
	for (int i=0;i<ntniv;i++) printf("corrLevelCol[%d] : %d\n",i,corrLevelColPtr[i]);
	for (int i=0;i<nd;i++) printf("Y[%d]         : %f\n",i,(double)Y_d[i*nsim]);
	for (int i=0;i<nd;i++) printf("CD[%d]         : %f\n",i,(double)CD_d[i]);
	for (int i=0;i<np;i++) printf("sigsq[%d]         : %f\n",i,(double)sigsquare_host[i*nsim]);
	printf("LAST Y[%d]        : %f\n",(nd-1)*nsim,(double)Y_d[(nd-1)*nsim]);
	printf("DEPASSEMENT 1 - LAST Y[%d]        : %f\n",(nd)*nsim,(double)Y_d[(nd)*nsim]);
#endif
	/* Check compatibility */

	if ( MAXND_WORK < nd ) {
		printf("Maximum number of progeny used by this analysis:%d [data.nd:%d]\n",MAXND_WORK,ndmax);
		exit(1);
	}
	/* Y */
	int size = nsim*nd;
	int err = cudaMalloc(&Y,size*sizeof(DT));
	if ( err ) {
		printf("Error allocation Y ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	safecall(cudaMemcpy(Y, Y_d, size*sizeof(DT), cudaMemcpyHostToDevice));

	/* CD */
	size = nd;
	safecall(cudaMalloc(&CD,size*sizeof(DT)));
	safecall(cudaMemcpy(CD, CD_d, size*sizeof(DT), cudaMemcpyHostToDevice));

	/* les effets fixes */
	size = ndmax*nLevelFix ;
	safecall(cudaMalloc(&contingence_fix,size*sizeof(DT)));
	safecall(cudaMemcpy(contingence_fix, xinc_d_fix, size*sizeof(DT), cudaMemcpyHostToDevice));

	safecall(cudaMallocHost(&contingence_fix_host,size*sizeof(DT)));
	for (int i=0;i<size;i++) {
		contingence_fix_host[i] = xinc_d_fix[i];
	}
    
	if ( nLevelVar > 0 ) {
		/* isigsq */
		size = np*nsim*nqtl ;
		safecall(cudaMalloc(&isigsq,size*sizeof(DT)));
		safecall(cudaMemcpy(isigsq,allVarFitted, size*sizeof(DT), cudaMemcpyHostToDevice));
	} 

	int ip=0;
	int subtotal=sizeFamilyNp[ip];
	int corIpKd_host[nd];
	for (int kd=0;kd<nd;kd++) {
		if ( kd < subtotal ) {
			corIpKd_host[kd]=ip;
		} else {
			ip++;
			subtotal+=sizeFamilyNp[ip];
			corIpKd_host[kd]=ip;
		}					
	}
	/* corIpKd */
	size = nd ;
	safecall(cudaMalloc(&corIpKd,size*sizeof(int)));
	safecall(cudaMemcpy(corIpKd, corIpKd_host, size*sizeof(int), cudaMemcpyHostToDevice));
	safecall(cudaGetLastError());

}


void QTLMapStructDeviceDataLinear::set_contingence_var(int ndmax, int ,int nbPositionsTest, int startPosition, DT* xinc_d_var,cudaStream_t * stream) {
	if ( nLevelVar > 0 ) {
		int size = nd*nLevelVar*nbPositionsTest*sizeof(DT) ;
		/* 1er appel initialisation du block memoire */
		if ( contingence_var == NULL ) {
			/* les effets variables a chaque position */
			safecall(cudaMalloc(&contingence_var,size));

		}

		/* rearrangement memoire */
		/* On peut stocker le pointeur et eviter une reallocation memoire du block temporaire pour le transfert */
		DT* n_contvar;
		safecall(cudaMallocHost(& n_contvar, size));

		for (int i=0;i<(nLevelVar*nbPositionsTest);i++) {
			for (int j=0;j<nd;j++) {
				n_contvar[j*(nLevelVar*nbPositionsTest)+i] = xinc_d_var[i*ndmax+j];
			}
		}

		/* */
		safecall(cudaMemcpyAsync(contingence_var, n_contvar, size, cudaMemcpyHostToDevice,*stream));
		safecall(cudaFreeHost(n_contvar));
	}
}

QTLMapStructDeviceDataLinear::~QTLMapStructDeviceDataLinear() {
}

void QTLMapStructDeviceDataLinear::releaseDevice(){
	safecall(cudaFree(contingence_fix));
	if ( contingence_var != NULL ) safecall(cudaFree(contingence_var));
	safecall(cudaFree(Y));
	safecall(cudaFree(CD));
	safecall(cudaFreeHost(contingence_fix_host));
	//	safecall(cudaFree(sizeFam));
	if (isigsq != NULL ) safecall(cudaFree(isigsq));
}


QTLMapStructDeviceWorkLinear::QTLMapStructDeviceWorkLinear() {
	XX = NULL ;
	triang = NULL ;
	rhs = NULL ;
}

void QTLMapStructDeviceWorkLinear::initResolution(bool hetero,const QTLMapStructDeviceDataLinear & data,DT *CD,DT * sigsquare_host,int * sizeFamily) {
	//printf("*********************** INIT WORK *********************************\n");

	/* XX */
	int size = hetero?SIZEMATSYM(data.nsim,data.npos,data.ntniv):SIZEMATSYM(1,data.npos,data.ntniv) ;

	if ( size > 0 ) safecall(cudaMalloc(&XX,size*sizeof(DT)));

#if(_CUDA_HOST_DEBUG_)
	DT * XX_HOST ;
	safecall(cudaMallocHost(&XX_HOST,size*sizeof(DT)));
	for (int i=0;i<size;i++) 
		XX_HOST[i]=0.0;
	safecall(cudaMemcpy(XX,XX_HOST, size*sizeof(DT), cudaMemcpyHostToDevice));
	safecall(cudaFreeHost(XX_HOST));
#endif

	/* triang */
	if ( size > 0 ) safecall(cudaMalloc(&triang,size*sizeof(DT)));

#if(_CUDA_HOST_DEBUG_)
	DT * triang_HOST ;
	safecall(cudaMallocHost(&triang_HOST,size*sizeof(DT)));
	for (int i=0;i<size;i++) 
		triang_HOST[i]=0.0;
	safecall(cudaMemcpy(triang,triang_HOST, size*sizeof(DT), cudaMemcpyHostToDevice));
	safecall(cudaFreeHost(triang_HOST));
#endif

	/* A_res */
	size = hetero?0:data.nLevelFix*data.nLevelFix;
	safecall(cudaMalloc(&A_res,size*sizeof(DT)));

#if(_CUDA_HOST_DEBUG_)
	if ( size > 0 ) {
		DT * A_res_HOST ;
		safecall(cudaMallocHost(&A_res_HOST,size*sizeof(DT)));
		for (int i=0;i<size;i++) 
			A_res_HOST[i]=0.0;
		safecall(cudaMemcpy(A_res,A_res_HOST, size*sizeof(DT), cudaMemcpyHostToDevice));
		safecall(cudaFreeHost(A_res_HOST));
	}
#endif

	/* RHS */
	size = data.nsim*data.npos*data.ntniv;
	if ( size > 0 ) safecall(cudaMalloc(&rhs,size*sizeof(DT)));

#if(_CUDA_HOST_DEBUG_)
	DT * rhs_HOST ;
	safecall(cudaMallocHost(&rhs_HOST,size*sizeof(DT)));

	for (int i=0;i<size;i++) 
		rhs_HOST[i]=0.0;

	safecall(cudaMemcpy(rhs,rhs_HOST, size*sizeof(DT), cudaMemcpyHostToDevice));
	safecall(cudaFreeHost(rhs_HOST));
#endif
}

QTLMapStructDeviceWorkLinear::~QTLMapStructDeviceWorkLinear() {

}

void QTLMapStructDeviceWorkLinear::releaseDeviceResolution(){
	safecall(cudaFree(A_res));
	if (rhs != NULL) safecall(cudaFree(rhs));
	if (XX != NULL) safecall(cudaFree(XX));
	if (triang != NULL) safecall(cudaFree(triang));
}


QTLMapStructDeviceSolutionLinear::QTLMapStructDeviceSolutionLinear() {
	bestim = NULL;
	osigsq = NULL;
	lrt    = NULL;
}

void QTLMapStructDeviceSolutionLinear::init(const QTLMapStructDeviceDataLinear & data) {
	/* bestim */
	int size = data.nsim*data.npos*data.ntniv*sizeof(DT);
	safecall(cudaMalloc(&bestim,size));

	/* osigsq */
	size = data.nsim*data.npos*data.np*sizeof(DT);
	safecall(cudaMalloc(&osigsq,size));

	/* lrt */
	if ( data.nqtl > 0) {
		size = data.nqtl*data.nsim*data.nposGlobal*sizeof(DT);
		safecall(cudaMalloc(&lrt,size));
	} 
}

QTLMapStructDeviceSolutionLinear::~QTLMapStructDeviceSolutionLinear() {

}


void QTLMapStructDeviceSolutionLinear::releaseDevice(){
	if ( bestim != NULL ) safecall(cudaFree(bestim));
	if ( osigsq != NULL ) safecall(cudaFree(osigsq));
	if ( lrt != NULL ) safecall(cudaFree(lrt));
}

template <class T>
void Utils<T>::printFloatDeviceArray1D(int dim1,int nb1,T* array1D) {
	int size = dim1;

	T* array1DHost ;
	int errMH = (int) cudaMallocHost(& array1DHost, size*sizeof(T));
	if ( errMH ) {
		printf("Error host allocation of printFloatDeviceArray1D::array1DHost ERR=%d \n",errMH);
		exit(1);
	}

	cudaError_t err = cudaMemcpy(array1DHost, array1D, size*sizeof(T), cudaMemcpyDeviceToHost);
	safecall(err);
	cout.precision(PRECISION_FLOTTANT);
	if ( err == 0 ) {
		cout << "****************************"<< endl;
		for (int i=0;i<nb1;i++ ) {
			cout << " " << array1DHost[i];
		}
		cout << endl <<"--"<< endl ;
		cout << "****************************"<< endl ;
	} else {
		cerr << "Error Memcpy :"<< err << endl ;
		exit(1);
	}
	safecall(cudaFreeHost(array1DHost));
}

template <class T>
void Utils<T>::printFloatDeviceArray2D(int dim1,int dim2,int nb1,int nb2,T* array2D) {
	int size = dim1*dim2;

	T* array2DHost ;
	int errMH = (int) cudaMallocHost(& array2DHost, size*sizeof(T));
	if ( errMH ) {
		printf("Error host allocation of printFloatDeviceArray2D::array2DHost ERR=%d \n",errMH);
		exit(1);
	}

	cudaError_t err = cudaMemcpy(array2DHost, array2D, size*sizeof(T), cudaMemcpyDeviceToHost);
	safecall(err);
	cout.precision(PRECISION_FLOTTANT);
	if ( err == 0 ) {
		cout << "****************************"<< endl ;
		for (int i=0;i<nb1;i++ ) {
			for (int j=0;j<nb2;j++ ) {
				cout << " " << array2DHost[dim1*j+i];
			}
			cout << endl ;
		}
		cout << endl <<"--"<< endl ;
		cout << "****************************"<< endl ;
	} else {
		cerr << "Error Memcpy :"<< err << endl ;
		exit(1);
	}
	safecall(cudaFreeHost(array2DHost));
}

template <class T>
void Utils<T>::printFloatDeviceArray3D(int dim1,int dim2,int dim3,int nb1,int nb2,int nb3,T* array3D) {
	int size = dim1*dim2*dim3;
	T* array3DHost ;
	int errMH = (int) cudaMallocHost(& array3DHost, size*sizeof(T));
	if ( errMH ) {
		printf("Error host allocation of printFloatDeviceArray3D::array3DHost ERR=%d \n",errMH);
		exit(1);
	}
 
	cudaError_t err = cudaMemcpy(array3DHost, array3D, size*sizeof(T), cudaMemcpyDeviceToHost);
	safecall(err);
	cout.precision(PRECISION_FLOTTANT);
	if ( err == 0 ) {
		cout << "****************************"<< endl ;
		for (int i=0;i<nb1;i++ ) {
			for (int j=0;j<nb2;j++ ) {
				for (int k=0;k<nb3 ;k++ ) {
					cout << " " << array3DHost[dim1*dim2*k+dim1*j+i];
				}
				cout << endl ;
			}
			cout << endl <<"--"<< endl ;
		}
		cout << "****************************"<< endl ;
	} else {
		cerr << "Error Memcpy :"<< err << endl ;
		exit(1);
	}
	safecall(cudaFreeHost(array3DHost));
}

template <class T>
void Utils<T>::printFloatDeviceArray4D(int dim1,int dim2,int dim3,int dim4,int nb1,int nb2,int nb3,int nb4,T* array4D) {
	int size = dim1*dim2*dim3*dim4;
	T* array4DHost ;
	int errMH = (int) cudaMallocHost(& array4DHost, size*sizeof(T));
	if ( errMH ) {
		printf("Error host allocation of printFloatDeviceArray4D::array4DHost ERR=%d \n",errMH);
		exit(1);
	}
	cout.precision(PRECISION_FLOTTANT);
	cudaError_t err = cudaMemcpy(array4DHost, array4D, size*sizeof(T), cudaMemcpyDeviceToHost);
	safecall(err);
	if ( err == 0 ) {
		cout << "****************************"<< endl ;
		for (int i=0;i<nb1;i++ ) {
			for (int j=0;j<nb2;j++ ) {
				for (int k=0;k<nb3 ;k++ ) {
					for (int l=0;l<nb4 ;l++ ) {
						cout << " " << array4DHost[l*dim1*dim2*dim3+dim1*dim2*k+dim1*j+i];
					}
					cout << endl ;
				}
				cout << endl << endl;
			}
			cout << endl << " --*****************************************************************************-- "<<endl;
		}
		cout << "****************************"<< endl ;
	} else {
		cerr << "Error Memcpy :"<< err << endl ;
		exit(1);
	}
	safecall(cudaFreeHost(array4DHost));
}


template <class T>
void Utils<T>::printFloatHostArray2D(int dim1,int dim2,int nb1,int nb2,T* array2D) {

}

template <class T>
void Utils<T>::printFloatHostArray3D(int dim1,int dim2,int dim3,int nb1,int nb2,int nb3,T* array3D) {

}

template <class T>
void Utils<T>::getArrayDeviceToHost(int nbBloc,int nbThread,int size,T * inDeviceArray, T *outHostArray,cudaStream_t * stream) {
	cudaError_t err ;
	if ( stream != NULL ) {
		err = cudaMemcpyAsync(outHostArray,inDeviceArray, size*sizeof(T), cudaMemcpyDeviceToHost,*stream);
	} else {
		err = cudaMemcpy(outHostArray,inDeviceArray, size*sizeof(T), cudaMemcpyDeviceToHost);
	}
	safecall(err);
	if ( err ) {
		printf("Error QTLMapStructDeviceWorkLinear::getArrayDeviceToHost = %d \n",err);
		exit(1);
	}
}



void printMatSymXXTriang(bool hetero,QTLMapStructDeviceDataLinear data,DT * w) {

	int sSim=hetero?data.nsim:1;
	sSim=1;
#if(!OPTIMIZE_MEMORY_MATSYM)
	int size = hetero?data.ntniv*data.ntniv*data.npos*data.nsim:data.ntniv*data.ntniv*data.npos;
#else
	int size = hetero?(CEIL(data.ntniv*(data.ntniv+1))/2)*data.npos*data.nsim:((data.ntniv*(data.ntniv+1))/2)*data.npos;
#endif	
	DT* array ;
	safecall(cudaMallocHost(& array, size*sizeof(DT)));
	safecall(cudaMemcpy(array, w, size*sizeof(DT), cudaMemcpyDeviceToHost));

	cout.precision(PRECISION_FLOTTANT);
	for (int isim=0;isim<sSim;isim++) {
		cout << endl << "************ ISIM:"<< isim << endl;
		for (int ipos=0;ipos<1;ipos++) {
			cout << " **> POS:"<< ipos << endl << "****"<< endl ;
			for (int iniv=0;iniv<data.ntniv;iniv++) {
				for (int jniv=0;jniv<data.ntniv;jniv++) {
					cout << array[GETACC(data,isim,ipos,iniv,jniv)]<<" ";
				}
				cout << endl ;
			}
		}
	}
	safecall(cudaFreeHost(array));
}


/**************************************************** MODEL METHODS ******************************************************************************************/


void QTLMapHomoscedasticModelCalcul::start_analysis(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work){
	/* calcul de XX' pour la partie des effets fixe a la position sur le host */
	int size = data.nLevelFix*data.nLevelFix;
	DT A_res_host[size];

	for (int i=0;i<data.nLevelFix;i++) {
		for (int j=i;j<data.nLevelFix;j++) {
			A_res_host[j*data.nLevelFix+i] = 0.0;
			for (int kd=0;kd<data.nd;kd++) {
				A_res_host[j*data.nLevelFix+i] += data.contingence_fix_host[i*data.ndmax+kd]*data.contingence_fix_host[j*data.ndmax+kd];
			}
		}
	}

	safecall(cudaMemcpy(work.A_res, A_res_host, size*sizeof(DT), cudaMemcpyHostToDevice));
}

__global__ void initialize_start_variance_H0(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,DT * varInDevice) {
	int isim = blockIdx.x * blockDim.x + threadIdx.x ;

	if ( isim < data.nsim ) {

		for (int ip=0;ip<data.np;ip++) {
			DT somyp = 0;
			int effp = 0;
			DT var   = 0;	


			for (int kd=0;kd<data.nd;kd++ ) {
				if ( ip == data.corIpKd[kd] )	{
					somyp = somyp + data.Y[isim+kd*data.nsim]*data.CD[kd];
					effp++;
				}
			}

			/* mean */
			DT mu = somyp / DT(effp);

			/* variance */
			for (int kd=0;kd<data.nd;kd++ ) {
				if ( ip == data.corIpKd[kd] ) {
					DT v = data.Y[isim+kd*data.nsim] - mu ;
					var = var + v*v;
				}
			}

			var = var / (DT(effp) - 1);
			varInDevice[isim+ip*data.nsim] = var;
		}
	}

}

/**
 * Initialize variance to start the model (classical variance estimator)
 * Calcul de XX pour les fixe a la position
 */
__global__ void initialize_start_variance(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,DT * varInDevice) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ;
	int isim = blockIdx.y ;	

	if ( ipos < data.npos && isim < data.nsim ) {
		for (int ip=0;ip<data.np;ip++) {
			int k = (data.nqtl-1)*data.nsim*data.np;
			varInDevice[ipos+isim*data.npos+ip*data.npos*data.nsim] = data.isigsq[isim+ip*data.nsim+k] ;
		}
	}
}

void QTLMapHeteroscedasticModelCalcul::start_analysis(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work){
	/* compute start variance for analysis under H0 */


	if ( data.nqtl == 0 ) {
		dim3 dimBlock(128);
		dim3 dimGrid(ceil(data.nsim / dimBlock.x )+1);
		initialize_start_variance_H0<<<dimGrid,dimBlock,0,stream>>>(data,work,varInDevice);

	} else {
		dim3 dimBlock(128);
		dim3 dimGrid(ceil(data.npos / dimBlock.x )+1,data.nsim);
		initialize_start_variance<<<dimGrid,dimBlock,0,stream>>>(data,work,varInDevice);

	}

	cudaThreadSynchronize();

#if(_CUDA_HOST_DEBUG_)
	cudaThreadSynchronize();
	printf("PARTIEL IVAR0\n------------POS=1,SIM=1,..,nsim => NP----------------------------------------------\n");
	Utils<DT>::printFloatDeviceArray3D(data.npos,data.nsim,data.np,data.npos,1,data.np,varInDevice);
#endif
}

void QTLMapHomoscedasticModelCalcul::calcul_XT_X_A(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	/* nothing to do */
}




__global__ void initialize_A_res(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,DT * varInDevice) {


	int ipos = blockIdx.x * blockDim.x + threadIdx.x ;
	int isim = blockIdx.y ;	

	if ( ipos < data.npos && isim < data.nsim ) {

		/* calcul de XX' des effets fixes a la position */
		for (int i=0;i<data.nLevelFix;i++) {
			for (int j=i;j<data.nLevelFix;j++) {
				DT v = 0;
				for (int kd=0;kd<data.nd;kd++) {
					v += data.contingence_fix[i*data.ndmax+kd]*data.contingence_fix[j*data.ndmax+kd]*(data.CD[kd]/varInDevice[ipos+isim*data.npos+data.corIpKd[kd]*data.npos*data.nsim]);
				}
				work.XX[GETACC(data,isim,ipos,constCorrIndexCol[j],constCorrIndexCol[i])] = v;
				//work.XX[GETACC(data,isim,ipos,constCorrIndexCol[i],constCorrIndexCol[j])] = v;
			}
		}
	}
}

void QTLMapHeteroscedasticModelCalcul::calcul_XT_X_A(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	dim3 dimBlock(128);
	dim3 dimGrid(ceil(data.npos / dimBlock.x )+1,data.nsim);
	initialize_A_res<<<dimGrid,dimBlock,0,stream>>>(data,work,varInDevice);
}

bool QTLMapHomoscedasticModelCalcul::convergenceOk(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceSolutionLinear solution) {
	return true;
}


__global__ void calcul_lrt_and_initialize_start_variance(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceSolutionLinear solution,DT * varInDevice,DT * lrtConv) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ;
	int isim = blockIdx.y ;

	if ( ipos < data.npos && isim < data.nsim ) {
		DT v = 0.0; 
		for (int ip=0;ip<data.np;ip++) {
			v += constSizeFam[ip] * (log(solution.osigsq[ipos+isim*data.npos+ip*data.npos*data.nsim]) - log(varInDevice[ipos+isim*data.npos+ip*data.npos*data.nsim]));
			varInDevice[ipos+isim*data.npos+ip*data.npos*data.nsim] = solution.osigsq[ipos+isim*data.npos+ip*data.npos*data.nsim];
		}
		lrtConv[ipos+isim*data.npos] = v;
	}
}

bool QTLMapHeteroscedasticModelCalcul::convergenceOk(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceSolutionLinear solution) {

	dim3 dimBlock(128);
	dim3 dimGrid(ceil(data.npos / dimBlock.x )+1,data.nsim);
	calcul_lrt_and_initialize_start_variance<<<dimGrid,dimBlock,0,stream>>>(data,solution,varInDevice,lrtConvergenceDevice);
	cudaThreadSynchronize();

	//
	int size=data.npos*data.nsim;
	DT  *lrt;
	//safecall(cudaMallocHost(&lrt,size*sizeof(DT)));
	lrt = new DT[size];

	safecall(cudaMemcpyAsync(lrt,lrtConvergenceDevice, size*sizeof(DT), cudaMemcpyDeviceToHost,stream));
	int countOk=0;

	for (int ipos=0;ipos<data.npos;ipos++) {
		for (int isim=0;isim<data.nsim;isim++) {
			int i = ipos + isim*data.npos ;
			if ( fabs(lastLrtConvergenceHost[i] - lrt[i]) < 0.5 )  countOk++;
			//			else
			//				cout << "ipos:"<< ipos << " isim:"<< isim << " "<< lrt[i] << " "<< lastLrtConvergenceHost[i] << " diff:" << fabs(lastLrtConvergenceHost[i] - lrt[i])<< endl ;
			lastLrtConvergenceHost[i] = lrt[i] ;
		}
	}

	delete [] lrt;
	//safecall(cudaFreeHost(lrt));

	if ( countOk ==  data.nsim*data.npos ) {
		cout << "************ COUNT CONV OK *****************" << endl ;
		return true;
	}
	
	cout << " **> COUNT CONV OK:" << countOk << "/"<< data.nsim*data.npos << endl ;

	return false;
}

/**********************************************************************************************************************************************
 * NFIX  : nombre de niveau fixe a la position (moyenne general, effet polygenique, effets de nuisances)
 * NVAR  : nombre de niveaux varaible a la position (effet qtl ou haplotype)
 * NTNIV : NFIX + NVAR 
 * NPOS  : nombre de position teste 
 * X     : matrice de contingence (animal par ligne / niveau exprime en colonne)
 * XT    : transpose de X
 * XT.X => Calcul independant de block A ( NFIX * NFIX operation ), B ( NFIX * NVAR * NPOS) , C (NVAR * NVAR/2 * NPOS)
 * 
 * Exemple : NFIX = 4, NVAR = 3   
 * 
 *              LEV FIX     POS1        POS2        POS3          POS4
 *           ( 1 1 0 0  0.5 0  0    0.4  0  0    0.3  0  0    0.2  0  0  )      NPOS = 4
 * TOTAL_X = ( 1 1 0 0  0.5 0  0    0.4  0  0    0.3  0  0    0.2  0  0  )
 *           ( 1 0 1 0  0  0.5 0     0  0.4  0    0  0.4  0    0  0.4  0 )    
 *           ( 1 0 1 0  0  0.5 0     0  0.4  0    0  0.4  0    0  0.4  0 )
 *           ( 1 0 0 1  0   0 0.5    0  0  0.4    0  0   0.4   0  0  0.4 )
 *           ( 1 0 0 1  0   0 0.5    0  0  0.4    0  0   0.4   0  0  0.4 )
 * 
 * pour POS = 1
 * 
 *     ( 1 1 0 0 0.5 0  0  )
 * X = ( 1 1 0 0 0.5 0  0  )
 *     ( 1 0 1 0  0 0.5 0  )     ND=6
 *     ( 1 0 1 0  0 0.5 0  )
 *     ( 1 0 0 1  0  0 0.5 )
 *     ( 1 0 0 1  0  0 0.5 )
 * 
 * 
 * XT.X = ( 1.1 1.2 1.3 1.4 1.5 1.6 1.7 )
 *        (     2.2 2.3 2.4 2.5 2.6 2.7 )
 *        (         3.3 3.4 3.5 3.6 3.7 )            ( 1.1 1.2 1.3 1.4 )         ( 1.5 1.6 1.7 )        ( 5.5 5.6 5.7 ) 
 *        (             4.4 4.5 4.6 4.7 )   avec A = (     2.2 2.3 2.4 )    B  = ( 2.5 2.6 2.7 )    C = (     6.6 6.7 )
 *        (                 5.5 5.6 5.7 )            (         3.3 3.4 )         ( 3.5 3.6 3.7 )        (         7.7 )
 *        (                     6.6 6.7 )            (             4.4 )         ( 4.5 4.6 4.7 )        
 *        (                         7.7 )
 * 
 * 
 *  - Calcul de A sur le HOST
 *  - Calcul de B sur le DEVICE
 *  - Calcul de C sur le DEVICE
 *  
 *  
 *  Calcul B :
 *  BLOCK à 2 dimensions : ivar => numero de la colonne dans la liste des niveaux variables
 *                         ifix => numero de la colonne dans la liste des niveaux fixes
 *                         
 *  
 *  Ordonnancement des resultats :
 *  [ifix] => |  1.5 pos1  | 1.6 pos1 |  1.7 pos1 |  1.5 pos2 |   | 1.6 pos2 |  1.7 pos2 ....
 *  
 */


__global__ void XT_X_B_homoscedastic(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {

	extern __shared__ DT FIXVALUE[];

	int ivar = blockIdx.x * blockDim.x + threadIdx.x ;

	if ( blockIdx.y < constNbLevelFix ) {
		int i_shared = threadIdx.x;
		while(i_shared < data.nd)
		{
			FIXVALUE[i_shared] = data.contingence_fix[blockIdx.y*data.ndmax+i_shared];
			i_shared += blockDim.x;
		}
		__syncthreads();
	}


	if ( ivar < constNbLevelVar*data.npos ) {
		DT v = 0;
		for (int kd=0;kd<data.nd;kd++ ) {
			v += FIXVALUE[kd]*data.contingence_var[kd*constNbLevelVar*data.npos+ivar];
		}

		//	work.B_res[blockIdx.y*constNbLevelVar*data.npos+ivar] =  v;
		/* niveau à une position */
		int iniv = ivar % constNbLevelVar ;
		/* position teste */
		int ipos = ivar / constNbLevelVar ;

		work.XX[GETACC(data,0,ipos,constCorrIndexCol[constNbLevelFix + iniv],constCorrIndexCol[blockIdx.y])] = v;
		//work.XX[GETACC(data,0,ipos,constCorrIndexCol[blockIdx.y],constCorrIndexCol[constNbLevelFix + iniv])] = v;
	}
}
void QTLMapHomoscedasticModelCalcul::calcul_XT_X_B(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {

	dim3 dimBlock(BLOCKDIMX);
	dim3 dimGrid (ceil(data.npos*data.nLevelVar / dimBlock.x) + 1 , (data.nLevelFix / dimBlock.y) + 1);
	XT_X_B_homoscedastic<<<dimGrid,dimBlock,data.nd*sizeof(DT),stream>>>(data,work);
	safecall(cudaGetLastError());
}



__global__ void XT_X_B_heteroscedastic(DT * varInDevice,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {

	extern __shared__ DT FIXVALUE[];

	int ivar = blockIdx.x * blockDim.x + threadIdx.x ;
	int isim = blockIdx.y ;
	int inivfix = blockIdx.z ;

	if ( inivfix < constNbLevelFix ) {
		int i_shared = threadIdx.x;
		while(i_shared < data.nd)
		{
			FIXVALUE[i_shared] = data.contingence_fix[inivfix*data.ndmax+i_shared];
			i_shared += blockDim.x;
		}
		__syncthreads();
	}


	if ( ivar < constNbLevelVar*data.npos && isim < data.nsim) {
		/* niveau à une position */
		int iniv = ivar % constNbLevelVar ;
		/* position teste */
		int ipos = ivar / constNbLevelVar ;

		DT v = 0.0;
		for (int kd=0;kd<data.nd;kd++ ) {
			v += FIXVALUE[kd]*data.contingence_var[kd*constNbLevelVar*data.npos+ivar]*(data.CD[kd] / varInDevice[ipos+isim*data.npos+data.corIpKd[kd]*data.nsim*data.npos]) ; 
		}

		work.XX[GETACC(data,isim,ipos,constCorrIndexCol[constNbLevelFix + iniv],constCorrIndexCol[inivfix])] = v;
		//work.XX[GETACC(data,isim,ipos,constCorrIndexCol[inivfix],constCorrIndexCol[constNbLevelFix + iniv])] = v;
	}
}

void QTLMapHeteroscedasticModelCalcul::calcul_XT_X_B(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {

	dim3 dimBlock(BLOCKDIMX);
	dim3 dimGrid (ceil(data.npos*data.nLevelVar / dimBlock.x) + 1 , data.nsim, data.nLevelFix);
	XT_X_B_heteroscedastic<<<dimGrid,dimBlock,data.nd*sizeof(DT),stream>>>(varInDevice,data,work);
}

/* calcul d'une valeur de la diagonale et des valeurs a droite de cette diagonale de XT.X   = colonne et ligne de niveau variable =  */
/*
 * 
 *     ( 5.5 5.6 5.7 )
 * C = (     6.6 6.7 )
 *     (         7.7 )
 *     
 *     
 *     stockage du resultat dans C_res sous la forme
 *     
 *            POS 1              POS2                   POS N
 *     | 5.5 | 6.6 | 7.7 | 5.5 | 6.6 | 7.7 | ... | 5.5 | 6.6 | 7.7 |
 *     puis
 *            POS 1     POS 2            POS N
 *     | 5.6 | 6.7 | X | 5.6 | 6.7 | X | ... | 5.6 | 6.7 | X |
 *     puis
 *       POS1              POS2               POS N
 *     [ 5.7 | X | X | 5.7 | X | X |.... | 5.7 | X | X |  
 * 
 */
__global__ void XT_X_C_homoscedastic(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	int ivar = blockIdx.x * blockDim.x + threadIdx.x ;

	/* ivar : la diagonale de XT . X */

	if ( ivar < constNbLevelVar*data.npos ) {	           
		/* iniv position entre 0 et ntlevVar */ 
		int nfix = constNbLevelFix ;
		int nvar = constNbLevelVar ;

		/* niveau à une position */
		int iniv = ivar % nvar ;
		/* position teste */
		int ipos = ivar / nvar ;

		/* initialisation */
		DT viniv[MAXND_WORK] ;
		DT v=0.0;

		for (int kd=0;kd<data.nd;kd++ ) {
			viniv[kd] = data.contingence_var[kd*nvar*data.npos+ivar];
			v += viniv[kd]*viniv[kd];		                                 
		}		

		/* calcul de la diagonal */

		/* les diagonales sont stockees en premieres...*/ 
		
		work.XX[GETACC(data,0,ipos,constCorrIndexCol[nfix + iniv],constCorrIndexCol[nfix + iniv])] = v;

		for (int jniv=1;jniv<(nvar-iniv);jniv++) {
			DT v = 0.0;	
			for (int kd=0;kd<data.nd;kd++ ) {
				v +=  viniv[kd]*data.contingence_var[kd*nvar*data.npos+(ivar+jniv)];
			}

			work.XX[GETACC(data,0,ipos,constCorrIndexCol[nfix + jniv + iniv],constCorrIndexCol[nfix + iniv])] = v;
			//work.XX[GETACC(data,0,ipos,constCorrIndexCol[nfix + iniv],constCorrIndexCol[nfix + jniv + iniv])] = v;
		}
	}
}


void QTLMapHomoscedasticModelCalcul::calcul_XT_X_C(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	dim3 dimBlock(BLOCKDIMX);
	dim3 dimGrid2 (ceil(data.npos*data.nLevelVar / dimBlock.x) + 1 , 1);
	XT_X_C_homoscedastic<<<dimGrid2,dimBlock,0,stream>>>(data,work);
}



__global__ void XT_X_C_heteroscedastic(DT * varInDevice,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	int ivar = blockIdx.x * blockDim.x + threadIdx.x ;
	int isim = blockIdx.y;

	/* ivar : la diagonale de XT . X */

	if ( ivar < constNbLevelVar*data.npos && isim  < data.nsim ) {	           
		/* iniv position entre 0 et ntlevVar */ 
		int nfix = constNbLevelFix ;
		int nvar = constNbLevelVar ;

		/* niveau à une position */
		int iniv = ivar % nvar ;
		/* position teste */
		int ipos = ivar / nvar ;

		/* initialisation */
		DT viniv[MAXND_WORK] ;
		DT v=0.0;

		//int scalesimXX = isim*BASE_SCALE_SYMMAT;
		v=0.0;
		for (int kd=0;kd<data.nd;kd++ ) {
			viniv[kd] = data.contingence_var[kd*nvar*data.npos+ivar];
			v += viniv[kd]*viniv[kd]*(data.CD[kd] / varInDevice[ipos+isim*data.npos+data.corIpKd[kd]*data.nsim*data.npos]);		                                 
		}		

		/* calcul de la diagonal */

		/* les diagonales sont stockees en premieres...*/ 

		//work.XX[ipos + constCorrIndexCol[nfix + iniv]*data.npos +constCorrIndexCol[nfix + iniv]*ntniv*data.npos + scalesimXX ] = v;
		work.XX[GETACC(data,isim,ipos,constCorrIndexCol[nfix + iniv],constCorrIndexCol[nfix + iniv])] = v;

		for (int jniv=1;jniv<(nvar-iniv);jniv++) {
			//scale +=  nvar*data.npos;
			DT v = 0.0;	

			for (int kd=0;kd<data.nd;kd++ ) {
				v +=  viniv[kd]*data.contingence_var[kd*nvar*data.npos+(ivar+jniv)]*(data.CD[kd] / varInDevice[ipos+isim*data.npos+data.corIpKd[kd]*data.nsim*data.npos]);
			}

			work.XX[GETACC(data,isim,ipos,constCorrIndexCol[nfix + jniv + iniv],constCorrIndexCol[nfix + iniv])] = v;
		//	work.XX[GETACC(data,isim,ipos,constCorrIndexCol[nfix + iniv],constCorrIndexCol[nfix + jniv + iniv])] = v;
		}
	}
}





void QTLMapHeteroscedasticModelCalcul::calcul_XT_X_C(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	dim3 dimBlock(BLOCKDIMX);
	dim3 dimGrid2 (ceil(data.npos*data.nLevelVar / dimBlock.x) + 1 , data.nsim);
	XT_X_C_heteroscedastic<<<dimGrid2,dimBlock,0,stream>>>(varInDevice,data,work);
}


/**
 *
 * Cholesky_Decomposition
 *
 *
 *
 */
__global__ void Cholesky_Decomposition_homoscedastic(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {

	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */

	DT * triang = work.triang ;
	int ntniv = data.ntniv;
	int nfix = constNbLevelFix;

	if ( ipos < data.npos ) {

		for (int i=0;i<ntniv;i++) {
			for (int j=i;j<ntniv;j++) {
				triang[GETACC(data,0,ipos,i,j)] = 0.0;
			}
		}
		/* a optimiser , ces valeurs sont redondantes dans les matrices XX pour chaque position  */
		for (int i=0;i<nfix;i++) {
			for (int j=i;j<nfix;j++) {
				work.XX[GETACC(data,0,ipos,constCorrIndexCol[j],constCorrIndexCol[i])] = work.A_res[j*nfix+i];
			}
		}

		for (int j=0;j<ntniv;j++) {
			//DT v = work.XX[ipos+j*data.npos +j*ntniv*data.npos]; //XX[j][j] ;
			DT v = work.XX[GETACC(data,0,ipos,j,j)]; //XX[j][j] ;

		for (int k=0;k<=j-1;k++) {
				// [j][j] = [j][j] - [k][j]*[k][j]
				//v -= triang[ipos+k*data.npos+j*ntniv*data.npos]*triang[ipos+k*data.npos+j*ntniv*data.npos];
				v -= triang[GETACC(data,0,ipos,k,j)]*triang[GETACC(data,0,ipos,k,j)];
			}
			//si estimable .. on ne met pas de structure vecseuil==> on utilise la diagonale de triangle pour savoir l effet est estimable
			//triang[ipos+j*data.npos +j*ntniv*data.npos] =  sqrt(v);
			triang[GETACC(data,0,ipos,j,j)] =  sqrt(v);
			//if ( triang[ipos+j*data.npos +j*ntniv*data.npos] != triang[ipos+j*data.npos +j*ntniv*data.npos] ) triang[ipos+j*data.npos +j*ntniv*data.npos] = 0.0;
			if ( triang[GETACC(data,0,ipos,j,j)] != triang[GETACC(data,0,ipos,j,j)] ) triang[GETACC(data,0,ipos,j,j)] = 0.0;
			if ( triang[GETACC(data,0,ipos,j,j)] > data.seuil_cho ) { 
				for (int i=j+1;i<ntniv;i++ ) {
					//triang[ipos+i*data.npos +j*ntniv*data.npos] = work.XX[ipos+i*data.npos +j*ntniv*data.npos];//XX[i][j];
					triang[GETACC(data,0,ipos,i,j)] = work.XX[GETACC(data,0,ipos,i,j)];//XX[i][j];
					for ( int k=0;k<=j-1;k++) {
						triang[GETACC(data,0,ipos,i,j)] -= (triang[GETACC(data,0,ipos,i,k)] * triang[GETACC(data,0,ipos,j,k)]);
					}
					triang[GETACC(data,0,ipos,i,j)] = triang[GETACC(data,0,ipos,i,j)] / triang[GETACC(data,0,ipos,j,j)];
					//triang[ipos+j*data.npos +i*ntniv*data.npos] = triang[ipos+i*data.npos +j*ntniv*data.npos] ;
				}
			}  
		}
	}

}


void QTLMapHomoscedasticModelCalcul::calcul_Cholesky_Decomposition(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {

	int nbblock_512 = ceil(data.npos / MAX_BLOCKDIM_512) + 1 ;
	Cholesky_Decomposition_homoscedastic<<<nbblock_512,MAX_BLOCKDIM_512,0,stream>>>(data,work);
}


__global__ void Cholesky_Decomposition_heterocedastic(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {

	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	int isim = blockIdx.y ;

	DT * triang = work.triang ;
    int ntniv = data.ntniv;

	if ( ipos < data.npos && isim < data.nsim ) {
		for (int i=0;i<ntniv;i++) {
			for (int j=i;j<ntniv;j++) {
				triang[GETACC(data,isim,ipos,i,j)] = 0.0;
			}
		}


		for (int j=0;j<ntniv;j++) {
			DT v = work.XX[GETACC(data,isim,ipos,j,j)]; //XX[j][j] ;

			for (int k=0;k<=j-1;k++) {
				// [j][j] = [j][j] - [k][j]*[k][j]
				v -= triang[GETACC(data,isim,ipos,j,k)]*triang[GETACC(data,isim,ipos,j,k)];
			}
			
			//si estimable .. on ne met pas de structure vecseuil==> on utilise la diagonale de triangle pour savoir l effet est estimable
			triang[GETACC(data,isim,ipos,j,j)] =  sqrt(v);
			if ( triang[GETACC(data,isim,ipos,j,j)] != triang[GETACC(data,isim,ipos,j,j)] ) triang[GETACC(data,isim,ipos,j,j)] = 0.0;

			if ( triang[GETACC(data,isim,ipos,j,j)] > data.seuil_cho ) { 
				for (int i=j+1;i<ntniv;i++ ) {
					triang[GETACC(data,isim,ipos,i,j)] = work.XX[GETACC(data,isim,ipos,i,j)];//XX[i][j];
					for ( int k=0;k<=j-1;k++) {
						triang[GETACC(data,isim,ipos,i,j)] -= (triang[GETACC(data,isim,ipos,i,k)] * triang[GETACC(data,isim,ipos,j,k)]);
					}
					triang[GETACC(data,isim,ipos,i,j)] = triang[GETACC(data,isim,ipos,i,j)] / triang[GETACC(data,isim,ipos,j,j)];
				}
			}
		}
	}

}


void QTLMapHeteroscedasticModelCalcul::calcul_Cholesky_Decomposition(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	dim3 dimBlock(BLOCKDIMX);
	dim3 dimGrid (ceil(data.npos / dimBlock.x) + 1 , data.nsim);
	Cholesky_Decomposition_heterocedastic<<<dimGrid,dimBlock,0,stream>>>(data,work);
}

/*
 * A optimiser : rhs pour les effet fixe a la position ne devrait prendre que nsim*nlevelfix et non pas nsim*nlevelfix*npos
 * 
 * 
 * logique de corrIndexCol :
 * 
 *     I=1,..,NTNIV | I=1,..,NFIX,NFIX+1,..,NVAR
 *     -------------------------------------------
 *     0            |            0
 *     1            |            J
 *     2            |            J+1
 *     3            |            .
 *     .            |            1
 *     .            |            2
 *     .            |            3
 *     NTNIV        |            NTNIV
 * 
 * 
 * 
 */
__global__ void Set_RHS_FIX_homoscedatic(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	int isim = blockIdx.y * blockDim.y + threadIdx.y ;// * blockDim.y + threadIdx.y ;

	int idNiv;
	if ( ipos < data.npos && isim < data.nsim ) {
		for (int iniv=0;iniv<constNbLevelFix;iniv++) {

			idNiv = constCorrIndexCol[iniv];

			int index = ipos + isim*data.npos + idNiv*data.nsim*data.npos ;
			work.rhs[index]=0;
			if (work.triang[GETACC(data,0,ipos,idNiv,idNiv)]>data.seuil_cho) {	
				for (int kd = 0;kd<data.nd;kd++ ) {
					work.rhs[index] += data.contingence_fix[iniv*data.ndmax + kd]*data.Y[isim+kd*data.nsim];
				}	
			}
		}
	}
}

__global__ void Set_RHS_VAR_homoscedatic(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	int ivar = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	int isim = blockIdx.y ;// * blockDim.y + threadIdx.y ;

	if ( ivar < constNbLevelVar*data.npos ) {	
		/* iniv position entre 0 et ntlevVar */ 
		
		/* niveau à une position */
		int iniv = ivar % constNbLevelVar ;
		/* position teste */
		int ipos = ivar / constNbLevelVar ;

		//for (int isim=0;isim<data.nsim;isim++) {
		int index = ipos + isim*data.npos + constCorrIndexCol[constNbLevelFix+iniv]*data.nsim*data.npos ;
		work.rhs[index]=0;
		if (work.triang[GETACC(data,0,ipos,constCorrIndexCol[constNbLevelFix+iniv],constCorrIndexCol[constNbLevelFix+iniv])]>data.seuil_cho) {
			for (int kd = 0;kd<data.nd;kd++ ) {
				work.rhs[index] += data.contingence_var[kd*constNbLevelVar*data.npos+ivar]*data.Y[isim+kd*data.nsim];
			}
		}
		//}
	}
}


void QTLMapHomoscedasticModelCalcul::calcul_RHS(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	dim3 dimBlock_Set_RHS_FIX(16,32);
	dim3 dimGrid_Set_RHS_FIX (ceil(data.npos*data.nLevelVar /dimBlock_Set_RHS_FIX.x ) + 1 ,ceil(data.nsim / dimBlock_Set_RHS_FIX.y )+1);
	Set_RHS_FIX_homoscedatic<<<dimGrid_Set_RHS_FIX,dimBlock_Set_RHS_FIX,0,stream>>>(data,work);
	dim3 dimGrid_Set_RHS_VAR (ceil(data.npos*data.nLevelVar / MAX_BLOCKDIM_64) + 1 ,data.nsim);
	Set_RHS_VAR_homoscedatic<<<dimGrid_Set_RHS_VAR,MAX_BLOCKDIM_64,0,stream>>>(data,work);
}


__global__ void Set_RHS_FIX_heteroscedastic(DT * varInDevice,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	int isim = blockIdx.y * blockDim.y + threadIdx.y ;// * blockDim.y + threadIdx.y ;

	int idNiv;
	if ( ipos < data.npos && isim < data.nsim ) {
		for (int iniv=0;iniv<constNbLevelFix;iniv++) {

			idNiv = constCorrIndexCol[iniv];

			int index = ipos + isim*data.npos + idNiv*data.nsim*data.npos ;
			work.rhs[index]=0;
			if (work.triang[GETACC(data,isim,ipos,idNiv,idNiv)]>data.seuil_cho) {
				for (int kd=0;kd<data.nd;kd++ ) {
					work.rhs[index] += data.contingence_fix[iniv*data.ndmax + kd]*data.Y[isim+kd*data.nsim]*(data.CD[kd] / varInDevice[ipos+isim*data.npos+data.corIpKd[kd]*data.nsim*data.npos]);
				}
			}
		}
	}
}

__global__ void Set_RHS_VAR_heteroscedastic(DT * varInDevice,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	int ivar = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	int isim = blockIdx.y ;// * blockDim.y + threadIdx.y ;

	if ( ivar < constNbLevelVar*data.npos ) {	
		/* iniv position entre 0 et ntlevVar */ 

		/* niveau à une position */
		int iniv = ivar % constNbLevelVar ;
		/* position teste */
		int ipos = ivar / constNbLevelVar ;

		int index = ipos + isim*data.npos + constCorrIndexCol[constNbLevelFix+iniv]*data.nsim*data.npos ;

		work.rhs[index]=0;
		
		if (work.triang[GETACC(data,isim,ipos,constCorrIndexCol[constNbLevelFix+iniv],constCorrIndexCol[constNbLevelFix+iniv])]>data.seuil_cho) {
			for (int kd = 0;kd<data.nd;kd++ ) {
				work.rhs[index] += data.contingence_var[kd*constNbLevelVar*data.npos+ivar]*data.Y[isim+kd*data.nsim]*(data.CD[kd] / varInDevice[ipos+isim*data.npos+data.corIpKd[kd]*data.nsim*data.npos]);
			}
		}
	}
}

void QTLMapHeteroscedasticModelCalcul::calcul_RHS(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	dim3 dimBlock_Set_RHS_FIX(16,32);
	dim3 dimGrid_Set_RHS_FIX (ceil(data.npos*data.nLevelVar /dimBlock_Set_RHS_FIX.x ) + 1 ,ceil(data.nsim / dimBlock_Set_RHS_FIX.y )+1);
	Set_RHS_FIX_heteroscedastic<<<dimGrid_Set_RHS_FIX,dimBlock_Set_RHS_FIX,0,stream>>>(varInDevice,data,work);
	dim3 dimGrid_Set_RHS_VAR (ceil(data.npos*data.nLevelVar / MAX_BLOCKDIM_64) + 1 ,data.nsim);
	Set_RHS_VAR_heteroscedastic<<<dimGrid_Set_RHS_VAR,MAX_BLOCKDIM_64,0,stream>>>(varInDevice,data,work);
}

/**
 *
 * Resolution LU (Partie descendante descendante )
 *
 *
 */
__global__ void Resolve_LU_homoscedastic(QTLMapStructDeviceDataLinear data, 
		QTLMapStructDeviceWorkLinear work,
		QTLMapStructDeviceSolutionLinear solution) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	int isim = blockIdx.y * blockDim.y + threadIdx.y ;

	int ntniv = data.ntniv ;
	DT temp[MAXNTNIV_LOCALTHREAD_MEMORY];

	/******* DESCENT LU *****/
	if ( ipos < data.npos && isim < data.nsim ) {
		//for (int isim=0;isim<data.nsim;isim++) {
		int scale = ipos + isim*data.npos ;
		for (int i=0;i<ntniv;i++) {
			temp[i] = 0.0;
			if (work.triang[GETACC(data,0,ipos,i,i)]>data.seuil_cho) {
				temp[i] = work.rhs[scale + i*data.nsim*data.npos];
				for (int j=i-1;j>=0;j--) {
					if (work.triang[GETACC(data,0,ipos,j,j)]>data.seuil_cho) {
						temp[i] = temp[i] - temp[j]*work.triang[GETACC(data,0,ipos,i,j)];
					}
				}
				temp[i] = temp[i] / work.triang[GETACC(data,0,ipos,i,i)];
			} 
		}

		for (int i=ntniv-1;i>=0;i--) {
			int indexI = scale + i*data.nsim*data.npos;
			solution.bestim[indexI] = 0.0 ;
			if (work.triang[GETACC(data,0,ipos,i,i)]>data.seuil_cho) {
				solution.bestim[indexI] = temp[i];
				for (int j=i+1;j<ntniv;j++) {
					if (work.triang[GETACC(data,0,ipos,j,j)]>data.seuil_cho) {
						int indexJ = scale + j*data.nsim*data.npos;
						solution.bestim[indexI] -= solution.bestim[indexJ]*work.triang[GETACC(data,0,ipos,i,j)];
					}
				}
				solution.bestim[indexI] = solution.bestim[indexI] / work.triang[GETACC(data,0,ipos,i,i)]; 
			} 
		}
		//}//fin for isim

	}
}


void QTLMapHomoscedasticModelCalcul::calcul_LU(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,QTLMapStructDeviceSolutionLinear solution) {
	/* Solution */
	dim3 Resolve_LU_dimBlock(16,32);
	dim3 Resolve_LU_dimGrid (ceil(data.npos/Resolve_LU_dimBlock.x)+1 , ceil(data.nsim / Resolve_LU_dimBlock.y)+1);
	Resolve_LU_homoscedastic<<<Resolve_LU_dimGrid,Resolve_LU_dimBlock,0,stream>>>(data,work,solution);
}


__global__ void Resolve_LU_heteroscedastic(QTLMapStructDeviceDataLinear data, 
		QTLMapStructDeviceWorkLinear work,
		QTLMapStructDeviceSolutionLinear solution) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	int isim = blockIdx.y * blockDim.y + threadIdx.y ;

	int ntniv = data.ntniv ;

	DT temp[MAXNTNIV_LOCALTHREAD_MEMORY];

	/******* DESCENT LU *****/
	if ( ipos < data.npos && isim < data.nsim ) {
		int scale = ipos + isim*data.npos ;
		for (int i=0;i<ntniv;i++) {
			temp[i] = 0.0;
			if (work.triang[GETACC(data,isim,ipos,i,i)]>data.seuil_cho) {
				temp[i] = work.rhs[scale + i*data.nsim*data.npos];
				for (int j=i-1;j>=0;j--) {
					if (work.triang[GETACC(data,isim,ipos,j,j)]>data.seuil_cho) {
						temp[i] = temp[i] - temp[j]*work.triang[GETACC(data,isim,ipos,j,i)];
					}
				}
				temp[i] = temp[i] / work.triang[GETACC(data,isim,ipos,i,i)];
			} 
		}

		for (int i=ntniv-1;i>=0;i--) {
			int indexI = scale + i*data.nsim*data.npos;
			solution.bestim[indexI] = 0.0 ;
			if (work.triang[GETACC(data,isim,ipos,i,i)]>data.seuil_cho) {
				solution.bestim[indexI] = temp[i];
				for (int j=i+1;j<ntniv;j++) {
					if (work.triang[GETACC(data,isim,ipos,j,j)]>data.seuil_cho) {
						int indexJ = scale + j*data.nsim*data.npos;
						solution.bestim[indexI] -= solution.bestim[indexJ]*work.triang[GETACC(data,isim,ipos,i,j)];
					}
				}
				solution.bestim[indexI] = solution.bestim[indexI] / work.triang[GETACC(data,isim,ipos,i,i)]; 
			} 
		}
	}
}



void QTLMapHeteroscedasticModelCalcul::calcul_LU(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,QTLMapStructDeviceSolutionLinear solution) {
	/* Solution */
	dim3 Resolve_LU_dimBlock(16,32);
	dim3 Resolve_LU_dimGrid (ceil(data.npos/Resolve_LU_dimBlock.x)+1 , ceil(data.nsim / Resolve_LU_dimBlock.y)+1);
	Resolve_LU_heteroscedastic<<<Resolve_LU_dimGrid,Resolve_LU_dimBlock,0,stream>>>(data,work,solution);
}




/**
 *
 * Set_SIGSQ
 * XB = Y - ( X' . Bestim )
 * SIGSQ = SUM(XB^2)
 */
__global__ void Set_SIGSQ_homoscedastic(QTLMapStructDeviceDataLinear data,
		QTLMapStructDeviceWorkLinear work,
		QTLMapStructDeviceSolutionLinear solution) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	int isim = blockIdx.y * blockDim.y + threadIdx.y ; /* simulation */

	int ntniv = data.ntniv ;

	if ( ipos < data.npos && isim < data.nsim ) {
		//	for (int isim=0;isim<data.nsim;isim++) {
		int scale = ipos + isim*data.npos ;
		DT xb = 0 ;

		for (int kd = 0;kd<data.nd;kd++ ) {
			DT v=0;

#pragma unroll 4
			for (int iniv=0;iniv<constNbLevelFix;iniv++) {
				v += data.contingence_fix[iniv*data.ndmax+kd]*solution.bestim[scale + constCorrIndexCol[iniv]*data.nsim*data.npos];
			}

			for (int iniv=constNbLevelFix;iniv<ntniv;iniv++) {	
				v += data.contingence_var[kd*constNbLevelVar*data.npos+ipos*constNbLevelVar+(iniv-constNbLevelFix)]*solution.bestim[scale + constCorrIndexCol[iniv]*data.nsim*data.npos];
			}

			v = data.Y[isim+kd*data.nsim] - v;
			xb += v*v ;
		}
		DT v = xb / data.nd ;

#pragma unroll 4
		for (int ip=0;ip<data.np;ip++) {
			solution.osigsq[ipos+isim*data.npos+ip*data.npos*data.nsim] = v ;
		}
	}
	//}
}

void QTLMapHomoscedasticModelCalcul::calcul_SIGSQ(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,QTLMapStructDeviceSolutionLinear solution) {
	dim3 dimBlock(16,32);
	dim3 dimGrid (ceil(data.npos/dimBlock.x)+1 , ceil(data.nsim / dimBlock.y)+1);					
	Set_SIGSQ_homoscedastic<<<dimGrid,dimBlock,0,stream>>>(data,work,solution);
}

/**
 *
 * Set_SIGSQ
 * XB = Y - ( X' . Bestim )
 * SIGSQ = SUM(XB^2)
 */
__global__ void Set_SIGSQ_heteroscedastic(QTLMapStructDeviceDataLinear data,
		QTLMapStructDeviceWorkLinear work,
		QTLMapStructDeviceSolutionLinear solution) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	int isim = blockIdx.y * blockDim.y + threadIdx.y  ; 

	int ntniv = data.ntniv ;

	if ( ipos < data.npos && isim < data.nsim ) {
		int scale = ipos + isim*data.npos ;

		for (int ip=0;ip<data.np;ip++) {
			DT xb = 0 ;
			for (int kd = 0;kd<data.nd;kd++ ) {
				if ( data.corIpKd[kd] == ip ) {
					DT v=0;

					for (int iniv=0;iniv<constNbLevelFix;iniv++) {
						v += data.contingence_fix[iniv*data.ndmax+kd]*solution.bestim[scale + constCorrIndexCol[iniv]*data.nsim*data.npos];
					}

					for (int iniv=constNbLevelFix;iniv<ntniv;iniv++) {	
						v += data.contingence_var[kd*constNbLevelVar*data.npos+ipos*constNbLevelVar+(iniv-constNbLevelFix)]*solution.bestim[scale + constCorrIndexCol[iniv]*data.nsim*data.npos];
					}

					v = data.Y[isim+kd*data.nsim] - v;
					xb += v*v*data.CD[kd]*data.CD[kd] ;
				}
			}
			solution.osigsq[ipos+isim*data.npos+ip*data.npos*data.nsim] = xb / constSizeFam[ip];
		}
	}
}


void QTLMapHeteroscedasticModelCalcul::calcul_SIGSQ(cudaStream_t stream,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,QTLMapStructDeviceSolutionLinear solution) {
	dim3 dimBlock(32,32);
	dim3 dimGrid (ceil(data.npos/dimBlock.x)+1 , ceil(data.nsim / dimBlock.y)+1);			
	Set_SIGSQ_heteroscedastic<<<dimGrid,dimBlock,0,stream>>>(data,work,solution);
}

/**
 *
 * Set_LRT
 *
 * 
 */
__global__ void Set_LRT(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,QTLMapStructDeviceSolutionLinear solution,int scalePosition) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */

	if ( data.nqtl > 0 ) {
		if ( ipos < data.npos ) {
			for (int isim=0;isim<data.nsim;isim++) {

				int k = ipos + isim * data.npos ;

				for (int iqtl=0;iqtl<data.nqtl;iqtl++) {
					DT v = 0.0;
					for (int ip=0;ip<data.np;ip++) {
						v += constSizeFam[ip] * (log(data.isigsq[isim+ip*data.nsim+(iqtl)*data.nsim*data.np]) - log(solution.osigsq[k+ip*data.npos*data.nsim]));
					}
					solution.lrt[scalePosition+ipos+isim*data.nposGlobal+iqtl*data.nposGlobal*data.nsim] = v;
				}
			}//fin for isim
		} 
	}
}

extern "C" void release_allocated_internal_sig_() 
{
	if (allVarFitted != NULL ) delete[] allVarFitted;
}

/**
 * - save residual variance at the maximum position of the likelihood. this information is useful to compute LRT
 */
void add_internal_sig(DT * osigsquare,int *maxPosition,int npos,int nsim,int np,int nqtl) {
	DT * allVarFitted_buf = NULL ;
	if ( allVarFitted ) {
		allVarFitted_buf = new DT [nqtl*np*nsim];
		for (int iq=0;iq<nqtl;iq++) {
			for (int isim=0;isim<nsim;isim++) {
				for (int ip=0;ip<np;ip++)
					allVarFitted_buf[isim+ip*nsim+iq*np*nsim] = allVarFitted[isim+ip*nsim+iq*np*nsim];
			}
		}
		delete[] allVarFitted;
		allVarFitted = NULL;
	}

	allVarFitted = new DT [(nqtl+1)*np*nsim];
	if ( nqtl >= 1 ) {
		for (int iq=0;iq<nqtl;iq++) {
			for (int isim=0;isim<nsim;isim++) {
				for (int ip=0;ip<np;ip++) {
					allVarFitted[isim+ip*nsim+iq*np*nsim] = allVarFitted_buf[isim+ip*nsim+iq*np*nsim];
				}
			}
		}
	}

	if (allVarFitted_buf) {
		delete [] allVarFitted_buf ;
		allVarFitted_buf = NULL;
	}

	/* update with the new values osig computed */
	for (int isim=0;isim<nsim;isim++) {
		if ( maxPosition[isim] >= 1 ) { /* index fortran */
			int ipos = maxPosition[isim] - 1;
			for (int ip=0;ip<np;ip++) {
				//			cout << osigsquare[ipos+isim*npos+ip*nsim*npos] << endl ;
				allVarFitted[isim+ip*nsim+nqtl*np*nsim] = osigsquare[ipos+isim*npos+ip*nsim*npos];	
			}
		}
	}
}



/*
  xinc  = matrice d'incidence
  nkd   = nombre de ligne
  ntniv = nombre de colonne
  seuil_choPtr = seuil d'estimabilite pour la solution

  column major order ===> offset = row + column*NUMROWS
  XINC ACCESS => ndmax*npos*iniv+npos*kd+ipos

  XINC(construction : level fixe puis levels variables)
  Mu MuPere1 MuPere2 MuMere1 MuMere2 QTLPere1Pos1 QTLPere2Pos1 QTLMere1Pos1 QTLMere2Pos1 ... QTLPere1PosN QTLPere2PosN QTLMere1PosN QTLMere2PosN

  nLevelFixPtr : 5 (Moyenne gen + 4 effet polygenique)
  nLevelVarPtr : 4 (4 effet qtl a estimer)
  corrLevelColPtr(1) = 1    moyenne generale
  corrLevelColPtr(2) = 6    effet qtl pere 1
  corrLevelColPtr(3) = 7    effet qtl pere 2
  corrLevelColPtr(4) = 8    effet qtl mere 1
  corrLevelColPtr(5) = 9    effet qtl mere 2
  corrLevelColPtr(6) = 2    effet polygenique pere 1
  corrLevelColPtr(7) = 3    effet polygenique pere 2
  corrLevelColPtr(8) = 4    effet polygenique mere 1
  corrLevelColPtr(9) = 5    effet polygenique mere 2
 */
extern "C" void cuda_homoscedatic_resolv_genome_(
		int *heteroscedastic_mode,            /* IN  : heteroscedastic = 1 , homoscedastic = other  */
		int *nqtlPtr,                         /* IN  : Number of QTL */
		DT  *sigsquare,                       /* IN  : variance residuelle sous les hypothese 0 à NQTL-1 tableau [NSIM][NP] */
		int *nLevelFixPtr,                    /* IN  : nombre de niveaux fixe a la position : moyenne genrale, polygenic, effet fixe, covariables */
		int *nLevelVarPtr,                    /* IN  : nombre de niveau variable a la position */
		int *corrLevelColPtr,                 /* IN  : Correspondance du niveau associe a une colonne de la matrice de contingence */
		void *work_cuda,                      /* IN  : structure de travail pour la fonction get_partialXinc */
		void (*get_partialXincFix)(void*,DT*),   /* IN  : Fonction qui donne les donnees fixe a la position de la matrice d incidences... */
		void (*get_partialXincVar)(void*,int*,int*,DT*),            /* IN  : Fonction qui donne les donnees variables a la position des matrices d incidences... */
		DT *Y_d,                              /* IN  : performance [nsim,nd] */
		DT *CD_d,                             /* IN  : cd of anim  [nd]      */
		int *nsimPtr,                         /* IN  : nombre de simulation */
		int *ndPtr,                           /* IN  : nombre d'individu total  : nombre de ligne reelle de XINC */
		int *nkdPtr,                          /* IN  : nombre d'individu pris en compte dans les matrices de contingence */
		int *npPtr,                           /* IN  : nombre de famille de pere */
		int *sizeFamilyNp,                    /* IN  : nombre de descendant par pere tableau [NP] */
		int *ntnivmaxPtr,                     /* IN  : nombre maximum de niveau : nombre de colonne reelle de XINC */
		int *npositionPtr,                    /* IN  : number of position estimation */
		double  *seuil_choPtr,                /* IN  : seuil pour considere grace a la decomposition de cholesky si un niveau est estimable */
		//		int    * vecsol,                      /* OUT : tableau de booleen de taille [NPOSITION,NTNIV]  pour l'estimabilite de chaque niveau */
		DT * bestim,                          /* OUT : tableau solution de l'estimation de chaque niveau (taille [NPOSITION,NSIM,NTNIV]) */
		DT * osigsquare,                      /* OUT : tableau des variances residuelles pour chaque familles de pere (taille [NPOSITION,NSIM,NP])  */
		DT * lrt,                             /* OUT : LRT (taille [NQTL,NPOSITION,NSIM])  */
		DT * maxLRT,                          /* OUT : Value of the Maximul LRT  [NQTL,NSIM] */
		int *maxPosition                      /* OUT : Index of the position maximum tableau [NSIM] */
) {
	bool hetero = (*heteroscedastic_mode == 1);

	QTLMapStructDeviceDataLinear      data ;
	QTLMapStructDeviceWorkLinear      work ;
	QTLMapStructDeviceSolutionLinear  solution ;

	safecall(cudaMemcpyToSymbol(constNbLevelFix, nLevelFixPtr, sizeof(int)));
	safecall(cudaMemcpyToSymbol(constNbLevelVar, nLevelVarPtr, sizeof(int)));
	safecall(cudaMemcpyToSymbol(constNSim, nsimPtr, sizeof(int)));

	if ( ( (*nLevelFixPtr) + (*nLevelVarPtr)) >= MAXNTNIV_LOCALTHREAD_MEMORY ) {
		cerr << "Can not initialised cuda structure => update constant MAXNTNIV_LOCALTHREAD_MEMORY :"<<  MAXNTNIV_LOCALTHREAD_MEMORY << "<=" <<  (*nLevelFixPtr + *nLevelVarPtr) << endl ;
		exit(1);
	}	

	safecall(cudaMemcpyToSymbol(constCorrIndexCol, corrLevelColPtr, (*nLevelFixPtr + *nLevelVarPtr)*sizeof(DT)));

	if ( (*npPtr) >=  MAXNP_WORK ) {
		cerr << "Can not initialised cuda structure => update constant MAXNP_WORK :"<<  MAXNP_WORK << "<=" <<  (*npPtr) << endl ;
		exit(1);
	}

	cudaMemcpyToSymbol(constSizeFam,sizeFamilyNp,(*npPtr)*sizeof(int));

#if(_CUDA_HOST_DEBUG_)
	printf("****************** INFO DEVICE ******************************\n");
	int nbDevice ;
	cudaGetDeviceCount(&nbDevice);
	printf("NB DEVICE     : %d\n",nbDevice);

	cudaDeviceProp  prop;

	for (int i=0;i<nbDevice;i++) {
		cudaGetDeviceProperties(&prop,i);
		printf(" ******* ===  %s  ===  *******\n",prop.name);
		printf(" ID DEVICE:%d\n",i);
		printf("  totalGlobalMem : %lu \n",prop.totalGlobalMem);
		printf("  sharedMemPerBlock : %lu \n",prop.sharedMemPerBlock);
		printf("  regsPerBlock : %d \n",prop.regsPerBlock);
		printf("  warpSize : %d \n",prop.warpSize);
		printf("  maxThreadsPerBlock : %d \n",prop.maxThreadsPerBlock);
		printf("  maxThreadsDim : %d,%d,%d \n",prop.maxThreadsDim[0],prop.maxThreadsDim[1],prop.maxThreadsDim[2]);
		printf("  maxGridSize : %d,%d,%d \n",prop.maxGridSize[0],prop.maxGridSize[1],prop.maxGridSize[2]);
		printf("  clockRate : %d \n",prop.clockRate);
		printf("  totalConstMem : %lu \n",prop.totalConstMem);
		printf("  multiProcessorCount : %d \n",prop.multiProcessorCount);
	}
	print_info_memory();
#endif

	int err ;

	DT *xinc_d_var,*xinc_d_fix;

	/* Initialisation des donnees fixes a la position de la matrice de contingence */
	safecall(cudaMallocHost(& xinc_d_fix, (*ndPtr)*((*nLevelFixPtr)*sizeof(DT))));
	get_partialXincFix(work_cuda,xinc_d_fix);

	data.init(nqtlPtr,sigsquare,xinc_d_fix,Y_d,CD_d,corrLevelColPtr,seuil_choPtr,ndPtr,nkdPtr,nsimPtr,npositionPtr,nLevelFixPtr,nLevelVarPtr,npPtr,sizeFamilyNp);

	int nbPosBlock;
	nbPosBlock = data.calculBlockPositionWorkSize(hetero);
	if ( nbPosBlock > *npositionPtr ) {
		nbPosBlock = *npositionPtr ;
	} 
	data.npos = nbPosBlock ;
	work.initResolution(hetero,data,CD_d,sigsquare,sizeFamilyNp) ;
	solution.init(data);

	/*
	 * 
	 *    Boucle des traitements par block
	 *    Utilisation de stream cuda
	 * 
	 * 
	 */
	/* Variables de travail pour recolter les resultats intermediaires */
	DT *local_osigsquare;
	DT *local_bestim;//[nbPosBlock*data.nsim*data.ntniv];

	safecall(cudaMallocHost(& local_osigsquare, nbPosBlock*data.nsim*data.np*sizeof(DT)));
	safecall(cudaMallocHost(& local_bestim, nbPosBlock*data.nsim*data.ntniv*sizeof(DT)));

	/* Creation des streams */
	int nstream=2;
	cudaStream_t stream[nstream];
	for (int i=0;i<nstream;i++) {
		cudaStreamCreate(&stream[i]);
	}
	int istream = 0 ;
	/* premier appel */

	int sizeBlockXincDVar = 0; /* taille en nombre de position a teste sur le groupe de liaison */
	int nbBlockXincDVar=2; /* On gere 2 block .l'un a remplir, l autre pour le calcul */
	int currentBlockMod = 0; 
	if ((*nLevelVarPtr) > 0) {
		/* Initialisation des donnees variables a la position de la matrice de contingence */
		sizeBlockXincDVar = (*ndPtr)*((*nLevelVarPtr)*nbPosBlock) ;
		safecall(cudaMallocHost(& xinc_d_var, nbBlockXincDVar*sizeBlockXincDVar*sizeof(DT)));
	}

	int lastPosition = 1;
	int nextPosition = nbPosBlock;
	get_partialXincVar(work_cuda,&lastPosition,&nextPosition,xinc_d_var);
	print_info_memory();

	QTLMapGenericModelCalcul *model ;
	if (hetero) {
		model = new QTLMapHeteroscedasticModelCalcul(data.npos,data.nsim,data.np);
	} else {
		model = new QTLMapHomoscedasticModelCalcul();
	}

	/*
	 * Initialisation pour benchmark
	 */	
#if(_CUDA_HOST_TIME_PROF_)
	cudaEvent_t start, stop;
	float elapsedTime;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
#endif

	for (int lastPosition = 1;lastPosition<=data.nposGlobal;lastPosition += nbPosBlock){   
#pragma omp sections
		{
#pragma omp section
			{
				int lastPositionNext = lastPosition + nbPosBlock ;
				if ( lastPositionNext <=data.nposGlobal ) {
					int nextPosition = lastPositionNext + nbPosBlock - 1;
					if ( nextPosition > data.nposGlobal) {
						nextPosition = data.nposGlobal;
					}

					int nextBlockMod = (currentBlockMod+1)%nbBlockXincDVar;
					get_partialXincVar(work_cuda,&lastPositionNext,&nextPosition,xinc_d_var+(nextBlockMod)*sizeBlockXincDVar);
				}
			}
#pragma omp section
			{
				nextPosition = lastPosition + nbPosBlock - 1;
				if ( nextPosition > data.nposGlobal) {
					nextPosition = data.nposGlobal;
				}
				data.npos = nextPosition - lastPosition + 1 ;

				cout << "   ** compute from position:"<< lastPosition << " at:"<< nextPosition << " (NPOS:"<< data.nposGlobal << ")   **" << endl << endl;

				/* Type de block utilise par les kernels */
				int nbblock_512 = ceil(data.npos / MAX_BLOCKDIM_512) + 1 ;
				int nbblock_64  = ceil(data.npos / MAX_BLOCKDIM_64) + 1 ;

				data.set_contingence_var(*ndPtr,*nLevelVarPtr,data.npos,0,xinc_d_var+currentBlockMod*sizeBlockXincDVar,&stream[istream]);//

				err = (int)cuInit(0);
				if ( err ) {
					printf(" cuInit error : %d \n",err);
					exit(1);
				}

				do {   /*  START IF CONVERGENCE IS OK */
					/* initializing model */
					model->start_analysis(stream[istream],data,work);
					cudaStreamSynchronize(stream[istream]);
#if(_CUDA_HOST_TIME_PROF_)
					cudaEventRecord(start, 0);
#endif		
					model->calcul_XT_X_A(stream[istream],data,work);
					/* on peut peut etre enlever cette synchronisation et la mettre plus tard */
					//cudaStreamSynchronize(stream[istream]);
#if(_CUDA_HOST_TIME_PROF_)
					cudaStreamSynchronize(stream[istream]);
					cudaEventRecord(stop, 0);
					cudaEventElapsedTime(&elapsedTime, start, stop);
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
					cout << "[BENCHMARK]  XT_X_A grid(x,y,z): elapsedtime:"<< elapsedTime<< endl ;
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif
#if(_CUDA_HOST_TIME_PROF_)
					cudaEventRecord(start, 0);
#endif				
					model->calcul_XT_X_B(stream[istream],data,work);
					/* on peut peut etre enlever cette synchronisation et la mettre plus tard */
					//cudaStreamSynchronize(stream[istream]);
					safecall(cudaGetLastError());

#if(_CUDA_HOST_TIME_PROF_)
					cudaStreamSynchronize(stream[istream]);
					cudaEventRecord(stop, 0);
					cudaEventElapsedTime(&elapsedTime, start, stop);
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
					cout << "[BENCHMARK]  XT_X_B grid(x,y,z): elapsedtime:"<< elapsedTime<< endl ;
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif


#if(_CUDA_HOST_TIME_PROF_)
					cudaEventRecord(start, 0);
#endif

					model->calcul_XT_X_C(stream[istream],data,work);
					cudaStreamSynchronize(stream[istream]);
					safecall(cudaGetLastError());

#if(_CUDA_HOST_TIME_PROF_)
					cudaEventRecord(stop, 0);
					cudaEventElapsedTime(&elapsedTime, start, stop);
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
					cout << "[BENCHMARK]  XT_X_C grid(x,y,z) elapsedtime:"<< elapsedTime<< endl ;
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif

#if(_CUDA_HOST_DEBUG_)
					printMatSymXXTriang(hetero,data,work.XX);
#endif

#if(_CUDA_HOST_TIME_PROF_)
					cudaEventRecord(start, 0);
#endif
					/* Cholesky decomp */
					model->calcul_Cholesky_Decomposition(stream[istream],data,work);
					cudaStreamSynchronize(stream[istream]);
					safecall(cudaGetLastError());
                     
				//	printMatSymXXTriang(hetero,data,work.triang);
					
#if(_CUDA_HOST_TIME_PROF_)
					cudaEventRecord(stop, 0);
					cudaEventElapsedTime(&elapsedTime, start, stop);
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
					cout << "[BENCHMARK]  Cholesky_Decomposition grid(x,y,z) elapsedtime:"<< elapsedTime<< endl ;
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif

#if(_CUDA_HOST_DEBUG_)
					printf("Cholesky decomp\n----------------------------------------------------------\n");
					printMatSymXXTriang(hetero,data,work.triang);
#endif

#if(_CUDA_HOST_TIME_PROF_)
					cudaEventRecord(start, 0);
#endif								
					model->calcul_RHS(stream[istream],data,work);
					cudaStreamSynchronize(stream[istream]);
					safecall(cudaGetLastError());

#if(_CUDA_HOST_TIME_PROF_)
					cudaEventRecord(stop, 0);
					cudaEventElapsedTime(&elapsedTime, start, stop);
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
					cout << "[BENCHMARK]  Set_RHS_FIX + Set_RHS_VAR grid(x,y,z):"<< " elapsedtime:"<< elapsedTime<< endl ;
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif


#if(_CUDA_HOST_DEBUG_)
					printf("RHS\n-------POS=1,SIM=nsim => NTNIV---------------------------------------------------\n");
					Utils<DT>::printFloatDeviceArray3D(data.npos,data.nsim,data.ntniv,10,1,data.ntniv,work.rhs);
#endif		

#if(_CUDA_HOST_TIME_PROF_)
					cudaEventRecord(start, 0);
#endif
					/* Solution */
					model->calcul_LU(stream[istream],data,work,solution);
					cudaStreamSynchronize(stream[istream]);
					safecall(cudaGetLastError());

#if(_CUDA_HOST_TIME_PROF_)
					cudaEventRecord(stop, 0);
					cudaEventElapsedTime(&elapsedTime, start, stop);
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
					cout << "[BENCHMARK]  Resolve_LU grid(x,y,z): elapsedtime:"<< elapsedTime<< endl ;
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif


#if(_CUDA_HOST_DEBUG_)
					printf("Resolution LU\n------POS=1,SIM=nsim => NTNIV ----------------------------------------------------\n");
					Utils<DT>::printFloatDeviceArray3D(data.npos,data.nsim,data.ntniv,4,1,data.ntniv,solution.bestim);
#endif

					/* Recuperation des resultats */
					Utils<DT>::getArrayDeviceToHost(nbblock_512,MAX_BLOCKDIM_512,data.npos*data.nsim*data.ntniv,solution.bestim,local_bestim,&stream[istream]);

#if(_CUDA_HOST_TIME_PROF_)
					cudaEventRecord(start, 0);
#endif
					/* Variances residuelles */
					model->calcul_SIGSQ(stream[istream],data,work,solution);
					/* on synchronise pour que local_bestim soit initialise */
					cudaStreamSynchronize(stream[istream]);

#if(_CUDA_HOST_TIME_PROF_)
					cudaEventRecord(stop, 0);
					cudaEventElapsedTime(&elapsedTime, start, stop);
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
					cout << "[BENCHMARK]  Set_SIGSQ grid(x,y,z)  elapsedtime:"<< elapsedTime<< endl ;
					cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif

					for (int ipos=0;ipos<data.npos;ipos++) {
						for (int isim=0;isim<data.nsim;isim++) {
							for (int iniv=0;iniv<data.ntniv;iniv++) {
								bestim[ipos+lastPosition-1+isim*data.nposGlobal+iniv*data.nsim*data.nposGlobal] = local_bestim[ipos+isim*data.npos+iniv*data.nsim*data.npos] ;
							}
						}
					}

					cudaStreamSynchronize(stream[istream]);
					safecall(cudaGetLastError());

#if(_CUDA_HOST_DEBUG_)
					cout << "SIGSQ------------POS=1,SIM=nsim => NP----------------------------------------------"<<endl ;
					Utils<DT>::printFloatDeviceArray3D(data.npos,data.nsim,data.np,data.npos,1,data.np,solution.osigsq);
#endif
					/* Recuperation des resultats */
					Utils<DT>::getArrayDeviceToHost(nbblock_512,MAX_BLOCKDIM_512,data.npos*data.nsim*data.np,solution.osigsq,local_osigsquare,&stream[istream]);



#if(_CUDA_HOST_TIME_PROF_)
					cudaEventRecord(start, 0);
#endif
					cudaStreamSynchronize(stream[istream]);
				} while ( ! model->convergenceOk(stream[istream],data,solution) ); /* FIN CONVERGENCE */

				if (data.nqtl > 0 ) {
					/* Calcul LRT */
					Set_LRT<<<nbblock_512,MAX_BLOCKDIM_512,0,stream[istream]>>>(data,work,solution,lastPosition-1);

#if(_CUDA_HOST_DEBUG_)
					printf("PARTIEL LRT\n------------POS=1,SIM=1,..,nsim => NP----------------------------------------------\n");
					cudaStreamSynchronize(stream[istream]);
					Utils<DT>::printFloatDeviceArray3D(data.nposGlobal,data.nsim,data.nqtl,data.nposGlobal,1,data.nqtl,solution.lrt);
#endif
					safecall(cudaGetLastError());
					cudaStreamSynchronize(stream[istream]);
				}


#if(_CUDA_HOST_TIME_PROF_)
				cudaEventRecord(stop, 0);
				cudaEventElapsedTime(&elapsedTime, start, stop);
				cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
				cout << "[BENCHMARK]  Set_LRT grid(x,y,z)  elapsedtime:"<< elapsedTime<< endl ;
				cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif



				for (int ipos=0;ipos<data.npos;ipos++) {
					for (int isim=0;isim<data.nsim;isim++) {
						for (int ip=0;ip<data.np;ip++) {
							osigsquare[ipos+lastPosition-1+isim*data.nposGlobal+ip*data.nsim*data.nposGlobal] = local_osigsquare[ipos+isim*data.npos+ip*data.nsim*data.npos];//[ip+isim*data.np+ipos*data.nsim*data.np] ;
						}
					}
				}
			}
		}
		istream = (istream+1)%nstream;
		currentBlockMod = (currentBlockMod+1)%nbBlockXincDVar ;

	}
	/* Fin boucle principale  */

	for (int i=0;i<nstream;i++) {
		cudaStreamDestroy(stream[i]);
	}


	work.releaseDeviceResolution();
	safecall(cudaFreeHost(xinc_d_fix));
	if ( data.nqtl>=1 ) safecall(cudaFreeHost(xinc_d_var));
	safecall(cudaFreeHost(local_osigsquare));
	safecall(cudaFreeHost(local_bestim));

	if ( data.nqtl>=1 ) {	
		/* Recuperation des LRTs */
		int nbblock_512_global = ceil(data.nposGlobal / MAX_BLOCKDIM_512) + 1 ;
		Utils<DT>::getArrayDeviceToHost(nbblock_512_global,MAX_BLOCKDIM_512,data.nqtl*data.nposGlobal*data.nsim,solution.lrt,lrt,NULL);

		for (int isim=0;isim<data.nsim;isim++) {
			for (int iq=0;iq<data.nqtl;iq++) {
				maxLRT[isim+iq*data.nsim]      = -999.9;
			}
			maxPosition[isim] = -1;

			for (int i=0;i<data.nposGlobal;i++) {
				if ( maxLRT[isim+(data.nqtl-1)*data.nsim] < lrt[i+isim*data.nposGlobal+(data.nqtl-1)*data.nposGlobal*data.nsim]) {
					for (int iq=0;iq<data.nqtl;iq++) {
						maxLRT[isim+iq*data.nsim] = lrt[i+isim*data.nposGlobal+iq*data.nposGlobal*data.nsim];
					}
					maxPosition[isim] = i+1 ; /* index pour fortran */
				}
			} // fin i
		} // fin isim

		add_internal_sig(osigsquare,maxPosition,data.nposGlobal,data.nsim,data.np,data.nqtl);		
	} else { /* residual variance under H0 */
		int *maxP;
		safecall(cudaMallocHost(&maxP,data.nsim*sizeof(DT)));
		for (int isim=0;isim<data.nsim;isim++)maxP[isim]=1;
		add_internal_sig(osigsquare,maxP,data.nposGlobal,data.nsim,data.np,data.nqtl);
		safecall(cudaFreeHost(maxP));
	}	
	delete model;

#if(_CUDA_HOST_TIME_PROF_)
	cudaEventDestroy(start);
	cudaEventDestroy(stop);
#endif

#if(_CUDA_HOST_DEBUG_)
	cout << "DATA RELEASED ON DEVICE" << endl ;
	data.releaseDevice();
#endif

#if(_CUDA_HOST_DEBUG_)
	cout << "SOLUTION RELEASED ON DEVICE" << endl ;
	solution.releaseDevice();
#endif


#if(_CUDA_HOST_DEBUG_)
	cout << " ** END ** " << endl ;
#endif

}
