/*
 *
 * O.Filangi : linearisation de la vraissemblance -  calcul simultané des position sur le génome
 * -deviceemu => emulation
 * nvcc -g -arch=sm_20 --cuda analyse_whole-genome.cu
 * debugging : ddd --debugger /usr/local/cuda/bin/cuda-gdb qtlmap
 * Debug   : nvcc -g -G -gencode=arch=compute_20,code=\"sm_20,compute_20\" --cuda analyse_whole-genome.cu
 * Release : nvcc -O4 --use_fast_math  -gencode=arch=compute_20,code=\"sm_20,compute_20\" --cuda analyse_whole-genome.cu
 * 
 * 
 * pour le profiling : /usr/local/cuda/computeprof/bin/computeprof
 */
#include <stdio.h>
#include <stdlib.h>
#include <cuda.h>
#include <iostream>
using namespace std;

#define safecall(call) do{\
		cudaError_t errERR = call ; \
		if (cudaSuccess != errERR){\
			fprintf(stderr, "cuda error at %s:%d, %s\n",\
					__FILE__, __LINE__, cudaGetErrorString(errERR));exit(1);\
		}\
}while(0) 


//#ifdef CUDA_SP
//#define DT float
//#else
//#define DT double
//#endif

#define DT double


/* Generate print information about sbu-results */
#define _HOST_DEBUG_ 0
/* Generate time comsuming information about sub process */
#define _HOST_TIME_PROF_ 1

/* number of thread available by block */
#define MAX_BLOCKDIM_512 512
#define MAX_BLOCKDIM_64 64
#define MAX_BLOCKDIM_32 32
#define BLOCKDIMX_XTX_C 96
#define BLOCKDIMX 128
#define MAXND_WORK 5000
#define MAXNP_WORK 100



#define MAXNTNIV_LOCALTHREAD_MEMORY 200

/* Definition des donnees en memoire constante sur gpu */

__constant__ int constNbLevelFix ;
__constant__ int constNbLevelVar ;
__constant__ int constNSim;
__constant__ int constCorrIndexCol[MAXNTNIV_LOCALTHREAD_MEMORY];
__constant__ int constSizeFam[MAXNP_WORK];




void print_info_memory() {
	size_t free,total;
	cuMemGetInfo(&free, &total) ;
	printf("===============> free: %f Mb (%lu bytes)   total:%f Mb (%lu bytes) \n",double(free)/(1024*1024),free,double(total)/(1024*1024),total);
}

class  QTLMapStructDeviceDataLinear {
public:
	//DT * contingence ; /* matrice de contingence les effets statiques sont dans les premieres colonne. */
	DT * contingence_fix ; /*                                      [NDMAX][NFIX]                         */
	DT * contingence_var ; /*                                      [NDMAX][NVAR][NPOS]                   */  
	DT * Y          ; /* Vecteur de Performance Y                  [NSIM][NDMAX]        */
	DT * CD         ; /* Vecteur des CD                            [NDMAX]        */
	int nqtl        ; /* hypothese courante */
	int np          ; /* nombre de famille de pere */
	int nsim        ; /* nombre de simulation */
	int nposGlobal  ; /* nombre de position teste */
	int npos        ; /* nombre de position teste par block */
	int ndmax       ; /* taille du tableau Y */
	int nd          ; /* nombre de descendants a prendre en compte */
	int nLevelFix   ; /* nombre d effet non dependant de la position */
	int nLevelVar   ; /* nombre d'effet dependant de la position */
	int ntniv       ; /* nLevelFix + nLevelVar */
	double seuil_cho    ; /* seuil pour l estimation des effets */
	DT * isigsq     ; /* variance residuelle des Hypothese < NQTL           [NSIM][NP] */
	int *corIpKd    ; /* *correspondance kd -> ip (indice du pere de la progeniture kd) */

public:

	QTLMapStructDeviceDataLinear();

	void init(   
			int * nqtlPtr,
			DT * sigsquare_host,
			DT * xinc_d_fix,
			DT * Y_d,
			DT * CD_d,
			int *corrLevelColPtr,
			double *seuil_choPtr,
			int * ndPtr,
			int * nkdPtr,
			int * nsimPtr,
			int * npositionGlobalPtr,
			//		int * npositionBlockPtr,
			int * nLevelFixPtr,
			int * nLevelVarPtr,
			int * npPtr,
			int * sizeFamilyNp);

	virtual ~QTLMapStructDeviceDataLinear();

	void set_contingence_var(int ndmax, int nLevelVar,int nbPositionsTest, int startPosition, DT* xinc_d_var,cudaStream_t * stream);

	void releaseDevice();

	size_t calculBlockPositionWorkSize() ;
} ;

class QTLMapStructDeviceWorkLinear {
public :
	DT * A_res      ; /* matrice d'incidence                                [NFIX][NFIX]*/

	DT * XX         ; /* matrice d'incidence                                [NPOS][NTNIV][NTNIV] */

	DT * triang     ; /* matrice temporaire de la decomposition Cholesky    [NPOS][NTNIV][NTNIV] */
	DT * rhs        ; /* Vecteur RHS                                        [NPOS][NSIM][NTNIV]  */

public :
	QTLMapStructDeviceWorkLinear();
	/* init de XX, triang, rhs */
	void initResolution(bool hetero,DT * xinc_d,const QTLMapStructDeviceDataLinear & data,DT *CD,DT * sigsquare_host,int * sizeFamily);

	virtual ~QTLMapStructDeviceWorkLinear();

	void releaseDeviceResolution();


	/* recuperation du tableau vecsol stoque sur le device */
	void getAllVecsolDevice(const QTLMapStructDeviceDataLinear & data,int nbBloc,int nbThread,int *vecsol) ;

} ;

class QTLMapStructDeviceSolutionLinear {
public:
	DT * osigsq     ; /* Vecteur des variances residuelles                  [NPOS][NSIM][NP]     */
	DT * bestim     ; /* Vecteur des solution des effets                    [NPOS][NSIM][NTNIV ] */
	DT * lrt        ; /* rapport de vraisemblance                           [NPOS][NSIM] */
public :

	QTLMapStructDeviceSolutionLinear();

	void init(const QTLMapStructDeviceDataLinear & data);

	virtual ~QTLMapStructDeviceSolutionLinear();

	void getAllMaxLrt(const QTLMapStructDeviceDataLinear & data,int nbBloc,int nbThread,DT * lrtMax,int * posMax);

	void releaseDevice();

} ;

size_t QTLMapStructDeviceDataLinear::calculBlockPositionWorkSize() {

	/* le calcul comprend 
	 * A_Res           : NFIX^2 
	 * XX              : NPOS*NTNIV*NTNIV
	 * Triang          : NPOS*NTNIV*NTNIV
	 * Rhs             : NPOS*NSIM*NTNIV
	 * OSIG            : NPOS*NSIM*NP
	 * BESTIM          : NPOS*NSIM*NTNIV
	 * LRT             : NPOSGLOBAL*NSIM
	 * CONTINGENCE_VAR : NPOS*NVAR*NPOS
	 * 
	 * 
	 * Y,CD et contingence_fix ont deja ete alloue 
	 * 
	 */

	size_t free,total;
	cuMemGetInfo(&free, &total) ;
	//  avec la version driver 3.7 on etait obliger de diviser par 1.8 pour que ca passe.....;
	free = free / 1.1 ;

	int num = ( nLevelFix*nLevelFix + nposGlobal*nsim + MAXND_WORK + np*nsim )*sizeof(DT);
	if ( free < num ) {
		cerr << "Not enough memory free:"<< free << " need:"<< num << endl ;
		exit(1);
	}
	size_t res = free - num;
	size_t denom = (2*ntniv*ntniv + 2*nsim*ntniv+ nsim*np + ndmax*nLevelVar)*sizeof(DT);
	res = ceil(res/denom)+1;
	cout << "**********************************************"<< endl;
	cout << "           **   Block NPOS :"<< res << " **"<< endl;
	cout << "**********************************************"<< endl;
	return res ;

} ;

template <class T>
class Utils {
public:
	static void printFloatDeviceArray1D(int dim1,int nb1,T* array1D);
	static void printFloatDeviceArray2D(int dim1,int dim2,int nb1,int nb2,T* array2D);
	static void printFloatDeviceArray3D(int dim1,int dim2,int dim3,int nb1,int nb2,int nb3,T* array3D);
	static void printFloatHostArray2D(int dim1,int dim2,int nb1,int nb2,T* array2D);
	static void printFloatHostArray3D(int dim1,int dim2,int dim3,int nb1,int nb2,int nb3,T* array3D);
	static void getArrayDeviceToHost(int nbBloc,int nbThread,int size,T * inDeviceArray, T *outHostArray,cudaStream_t * stream);
};

/*******************************************************************************************************************************************/

QTLMapStructDeviceDataLinear::QTLMapStructDeviceDataLinear() {
	contingence_fix = NULL ; 
	contingence_var = NULL ;
	Y = NULL;
	CD = NULL;
	isigsq = NULL;
}


void QTLMapStructDeviceDataLinear::init(
		int * nqtlPtr,
		DT * sigsquare_host,
		DT * xinc_d_fix,
		DT * Y_d,
		DT * CD_d,
		int *corrLevelColPtr,
		double *seuil_choPtr,
		int * ndPtr,
		int * nkdPtr,
		int * nsimPtr,
		int * npositionGlobalPtr,
		//		int * npositionBlockPtr,
		int * nLevelFixPtr,
		int * nLevelVarPtr,
		int * npPtr,
		int * sizeFamilyNp) {

	seuil_cho   = *seuil_choPtr;
	seuil_cho   = sqrt(seuil_cho);
	ndmax       = *ndPtr       ; /* taille reelle de la matrice de contingence : lignes */
	nd          = *nkdPtr      ; /* taille de la population pris en compte */
	nsim        = *nsimPtr     ;
	npos        = 0;
	nposGlobal  = *npositionGlobalPtr;
	nLevelFix   = *nLevelFixPtr ;
	nLevelVar   = *nLevelVarPtr;
	np          = *npPtr;
	ntniv       = nLevelFix + nLevelVar ;
	nqtl        = *nqtlPtr;

#if(_HOST_DEBUG_)
	printf("*********************** INIT DATA *********************************\n");
	printf("seuil_cho  SQ : %f\n",seuil_cho);
	printf("ndmax         : %d\n",ndmax);
	printf("nd            : %d\n",nd);
	printf("nsim          : %d\n",nsim);
	printf("np            : %d\n",np);
	printf("nposGlobal    : %d\n",nposGlobal);
	printf("npos          : %d\n",npos);
	printf("nLevelFix     : %d\n",nLevelFix);
	printf("nLevelVar     : %d\n",nLevelVar);
	for (int i=0;i<np;i++) printf("sizeFamilyNp[%d] : %d\n",i,sizeFamilyNp[i]);
	for (int i=0;i<ntniv;i++) printf("corrLevelCol[%d] : %d\n",i,corrLevelColPtr[i]);
	for (int i=0;i<10;i++) printf("Y[%d]         : %f\n",i,(double)Y_d[i*nsim]);
	for (int i=0;i<10;i++) printf("CD[%d]         : %f\n",i,(double)CD_d[i]);
	for (int i=0;i<np;i++) printf("sigsq[%d]         : %f\n",i,(double)sigsquare_host[i*nsim]);
	
	printf("LAST Y[%d]        : %f\n",(nd-1)*nsim,(double)Y_d[(nd-1)*nsim]);
	printf("DEPASSEMENT 1 - LAST Y[%d]        : %f\n",(nd)*nsim,(double)Y_d[(nd)*nsim]);
#endif
	/* Check compatibility */

	if ( MAXND_WORK < ndmax ) {
		printf("Maximum number of progeny used by this analysis:%d [data.nd:%d]\n",MAXND_WORK,ndmax);
		exit(1);
	}
	/* Y */
	int size = nsim * ndmax * sizeof(DT);
	int err = cudaMalloc(&Y,size);
	if ( err ) {
		printf("Error allocation Y ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	safecall(cudaMemcpy(Y, Y_d, size, cudaMemcpyHostToDevice));

	/* CD */
	size = ndmax * sizeof(DT);
	safecall(cudaMalloc(&CD,size));
	safecall(cudaMemcpy(CD, CD_d, size, cudaMemcpyHostToDevice));

	/* les effets fixes */
	size = ndmax*nLevelFix*sizeof(DT) ;
	safecall(cudaMalloc(&contingence_fix,size));
	safecall(cudaMemcpy(contingence_fix, xinc_d_fix, size, cudaMemcpyHostToDevice));

	/* ----------------------------------------------------------------------------------------------------------*/


	/* sizeFamilyNp */
//	size = np*sizeof(int) ;
//	safecall(cudaMalloc(&sizeFam,size));
//	safecall(cudaMemcpy(sizeFam, sizeFamilyNp, size, cudaMemcpyHostToDevice));

	if ( nLevelVar > 0 ) {
		/* isigsq */
		size = np*nsim*sizeof(DT) ;
		safecall(cudaMalloc(&isigsq,size));
		safecall(cudaMemcpy(isigsq, sigsquare_host, size, cudaMemcpyHostToDevice));
	}


	int ip=0;
	int subtotal=sizeFamilyNp[ip];
	int corIpKd_host[nd];
	for (int kd=0;kd<nd;kd++) {
		if ( kd < subtotal ) {
			corIpKd_host[kd]=ip;
		} else {
			ip++;
			subtotal+=sizeFamilyNp[ip];
			corIpKd_host[kd]=ip;
		}					
	}
	/* corIpKd */
	size = nd*sizeof(int) ;
	safecall(cudaMalloc(&corIpKd,size));

	safecall(cudaMemcpy(corIpKd, corIpKd_host, size, cudaMemcpyHostToDevice));

	safecall(cudaGetLastError());

}


void QTLMapStructDeviceDataLinear::set_contingence_var(int ndmax, int ,int nbPositionsTest, int startPosition, DT* xinc_d_var,cudaStream_t * stream) {
	if ( nLevelVar > 0 ) {
		int size = ndmax*nLevelVar*nbPositionsTest*sizeof(DT) ;
		/* 1er appel initialisation du block memoire */
		if ( contingence_var == NULL ) {
			/* les effets variables a chaque position */
			safecall(cudaMalloc(&contingence_var,size));

		}

		/* rearrangement memoire */
		/* On peut stocker le pointeur et eviter une reallocation memoire du block temporaire pour le transfert */
		DT* n_contvar;
		safecall(cudaMallocHost(& n_contvar, size));

		for (int i=0;i<(nLevelVar*nbPositionsTest);i++) {
			for (int j=0;j<nd;j++) {
				n_contvar[j*(nLevelVar*nbPositionsTest)+i] = xinc_d_var[i*ndmax+j];
			}
		}
		/* */
		safecall(cudaMemcpyAsync(contingence_var, n_contvar, size, cudaMemcpyHostToDevice,*stream));
		safecall(cudaFreeHost(n_contvar));
	}
}

QTLMapStructDeviceDataLinear::~QTLMapStructDeviceDataLinear() {
}

void QTLMapStructDeviceDataLinear::releaseDevice(){
	safecall(cudaFree(contingence_fix));
	if ( contingence_var != NULL ) safecall(cudaFree(contingence_var));
	safecall(cudaFree(Y));
	safecall(cudaFree(CD));
//	safecall(cudaFree(sizeFam));
	if (isigsq != NULL ) safecall(cudaFree(isigsq));
}


QTLMapStructDeviceWorkLinear::QTLMapStructDeviceWorkLinear() {
	XX = NULL ;
	triang = NULL ;
	rhs = NULL ;
}

void QTLMapStructDeviceWorkLinear::initResolution(bool hetero,DT * xinc_d_fix,const QTLMapStructDeviceDataLinear & data,DT *CD,DT * sigsquare_host,int * sizeFamily) {
	//printf("*********************** INIT WORK *********************************\n");

	/* XX */
	int size = data.ntniv*data.ntniv*data.npos*sizeof(DT) ;
	if ( size > 0 ) safecall(cudaMalloc(&XX,size));

	/* triang */
	if ( size > 0 ) safecall(cudaMalloc(&triang,size));

	/* A_res */
	size = data.nLevelFix*data.nLevelFix*sizeof(DT);
	safecall(cudaMalloc(&A_res,size));

	DT A_res_host[data.nLevelFix*data.nLevelFix];

	//	printf("******* A_RES ****** \n");
	/* Initialisation de la partie A de XX sur le host*/
	int ip=0;
	int subtotal=sizeFamily[ip];
	int isim=0;
	for (int i=0;i<data.nLevelFix;i++) {
		for (int j=i;j<data.nLevelFix;j++) {
			A_res_host[j*data.nLevelFix+i] = 0.0;
			if (hetero) {
				for (int kd=0;kd<data.nd;kd++) {
					if ( kd >= subtotal ) {
						ip++;
						subtotal+=sizeFamily[ip];
					}
					A_res_host[j*data.nLevelFix+i] += xinc_d_fix[i*data.ndmax+kd]*xinc_d_fix[j*data.ndmax+kd]*(CD[kd] / sigsquare_host[isim+ip*data.nsim]);
				}
			} else {
				for (int kd=0;kd<data.nd;kd++) {
					A_res_host[j*data.nLevelFix+i] += xinc_d_fix[i*data.ndmax+kd]*xinc_d_fix[j*data.ndmax+kd];
				}
			}
		}
	}
 
	safecall(cudaMemcpy(A_res, A_res_host, size, cudaMemcpyHostToDevice));

	/* RHS */
	size = data.nsim*data.npos*data.ntniv*sizeof(DT);
	if ( size > 0 ) safecall(cudaMalloc(&rhs,size));
}

QTLMapStructDeviceWorkLinear::~QTLMapStructDeviceWorkLinear() {

}

void QTLMapStructDeviceWorkLinear::releaseDeviceResolution(){
	safecall(cudaFree(A_res));
	if (rhs != NULL) safecall(cudaFree(rhs));
	if (XX != NULL) safecall(cudaFree(XX));
	if (triang != NULL) safecall(cudaFree(triang));
}

__global__ void calcul_vecsol(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,int *vec_device) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */

	if (ipos < data.npos ) {
		for(int i=0;i<data.ntniv;i++) {
			vec_device[ipos+i*data.npos] = (work.triang[ipos+i*data.npos+i*data.npos*data.ntniv] > data.seuil_cho );
		}
	}

}

void QTLMapStructDeviceWorkLinear::getAllVecsolDevice(const QTLMapStructDeviceDataLinear & data,int nbBloc,int nbThread,int * vecsol) {

	int * vec_device;
	int size = data.npos*data.ntniv*sizeof(int);
	safecall(cudaMalloc(&vec_device,size));

	calcul_vecsol<<<nbBloc,nbThread>>>(data,*this,vec_device);
	cudaThreadSynchronize();
	safecall(cudaGetLastError());
	safecall(cudaMemcpy(vecsol, vec_device, size, cudaMemcpyDeviceToHost));
	safecall(cudaFree(vec_device));
}

QTLMapStructDeviceSolutionLinear::QTLMapStructDeviceSolutionLinear() {
	bestim = NULL;
	osigsq = NULL;
	lrt    = NULL;
}

void QTLMapStructDeviceSolutionLinear::init(const QTLMapStructDeviceDataLinear & data) {
	//printf("*********************** INIT SOLUTION *********************************\n");
	/* bestim */
	int size = data.nsim*data.npos*data.ntniv*sizeof(DT);
	if (size>0) safecall(cudaMalloc(&bestim,size));

	/* osigsq */
	size = data.nsim*data.npos*data.np*sizeof(DT);
	if (size>0) safecall(cudaMalloc(&osigsq,size));

	/* lrt */
	size = data.nsim*data.nposGlobal*sizeof(DT);
	if (size>0) safecall(cudaMalloc(&lrt,size));
}

QTLMapStructDeviceSolutionLinear::~QTLMapStructDeviceSolutionLinear() {

}


/**
 *
 * set_maxLRT
 *
 *
 */
__global__ void set_maxLRT(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceSolutionLinear solution,DT* lrtMax,int * posMax) {
	int isim = blockIdx.x * blockDim.x + threadIdx.x ; /* index de la simul  */

	if ( isim < data.nsim ) {
		DT  mL = -999.9;
		int pMax;
		for (int ipos=0;ipos<data.nposGlobal;ipos++) {
			if (mL < solution.lrt[ipos+isim*data.nposGlobal]) {
				mL   = solution.lrt[ipos+isim*data.nposGlobal];
				pMax = ipos+1;
			}
		}
		lrtMax[isim] = mL;
		posMax[isim] = pMax;
	}
}


void QTLMapStructDeviceSolutionLinear::getAllMaxLrt(const QTLMapStructDeviceDataLinear & data,int nbBloc,int nbThread,DT * lrtMax,int * posMax) {

	int * posMax_device;
	DT  * lrtMax_device;
	int size = data.nsim ;

	safecall(cudaMalloc(&posMax_device,size*sizeof(int)));
	safecall(cudaMalloc(&lrtMax_device,size*sizeof(DT)));

	set_maxLRT<<<nbBloc,nbThread>>>(data,*this,lrtMax_device,posMax_device);
	cudaThreadSynchronize();
	safecall(cudaGetLastError());

	safecall(cudaMemcpy(lrtMax, lrtMax_device, size*sizeof(DT), cudaMemcpyDeviceToHost));
	safecall(cudaMemcpy(posMax, posMax_device, size*sizeof(int), cudaMemcpyDeviceToHost));

	//	cout << lrtMax[0] << endl ;
	//	cout << posMax[0] << endl ;

	safecall(cudaFree(lrtMax_device));
	safecall(cudaFree(posMax_device));

}



void QTLMapStructDeviceSolutionLinear::releaseDevice(){
	if ( bestim != NULL ) safecall(cudaFree(bestim));
	if ( osigsq != NULL ) safecall(cudaFree(osigsq));
	if ( lrt != NULL ) safecall(cudaFree(lrt));
}

template <class T>
void Utils<T>::printFloatDeviceArray1D(int dim1,int nb1,T* array1D) {
	int size = dim1;

	T* array1DHost ;
	int errMH = (int) cudaMallocHost(& array1DHost, size*sizeof(T));
	if ( errMH ) {
		printf("Error host allocation of printFloatDeviceArray1D::array1DHost ERR=%d \n",errMH);
		exit(1);
	}

	cudaError_t err = cudaMemcpy(array1DHost, array1D, size*sizeof(T), cudaMemcpyDeviceToHost);
	safecall(err);
	if ( err == 0 ) {
		printf("****************************\n");
		for (int i=0;i<nb1;i++ ) {
			printf(" %f ",(float)array1DHost[i]);
		}
		printf("\n -- \n");
		printf("****************************\n");
	} else {
		printf("Error Memcpy %d\n",err);
		exit(1);
	}
	safecall(cudaFreeHost(array1DHost));
}

template <class T>
void Utils<T>::printFloatDeviceArray2D(int dim1,int dim2,int nb1,int nb2,T* array2D) {
	int size = dim1*dim2;

	T* array2DHost ;
	int errMH = (int) cudaMallocHost(& array2DHost, size*sizeof(T));
	if ( errMH ) {
		printf("Error host allocation of printFloatDeviceArray2D::array2DHost ERR=%d \n",errMH);
		exit(1);
	}

	cudaError_t err = cudaMemcpy(array2DHost, array2D, size*sizeof(T), cudaMemcpyDeviceToHost);
	safecall(err);
	if ( err == 0 ) {
		printf("****************************\n");
		for (int i=0;i<nb1;i++ ) {
			for (int j=0;j<nb2;j++ ) {
				printf(" %f ",(float)array2DHost[dim1*j+i]);
			}
			printf("\n");
		}
		printf("\n -- \n");
		printf("****************************\n");
	} else {
		printf("Error Memcpy %d\n",err);
		exit(1);
	}
	safecall(cudaFreeHost(array2DHost));
}

template <class T>
void Utils<T>::printFloatDeviceArray3D(int dim1,int dim2,int dim3,int nb1,int nb2,int nb3,T* array3D) {
	int size = dim1*dim2*dim3;
	T* array3DHost ;
	int errMH = (int) cudaMallocHost(& array3DHost, size*sizeof(T));
	if ( errMH ) {
		printf("Error host allocation of printFloatDeviceArray3D::array3DHost ERR=%d \n",errMH);
		exit(1);
	}

	cudaError_t err = cudaMemcpy(array3DHost, array3D, size*sizeof(T), cudaMemcpyDeviceToHost);
	safecall(err);
	if ( err == 0 ) {
		printf("****************************\n");
		for (int i=0;i<nb1;i++ ) {
			for (int j=0;j<nb2;j++ ) {
				for (int k=0;k<nb3 ;k++ ) {
					printf(" %f ",(float)array3DHost[dim1*dim2*k+dim1*j+i]);
				}
				printf("\n");
			}
			printf("\n -- \n");
		}
		printf("****************************\n");
	} else {
		printf("Error Memcpy %d\n",err);
		exit(1);
	}
	safecall(cudaFreeHost(array3DHost));
}

template <class T>
void Utils<T>::printFloatHostArray2D(int dim1,int dim2,int nb1,int nb2,T* array2D) {

}

template <class T>
void Utils<T>::printFloatHostArray3D(int dim1,int dim2,int dim3,int nb1,int nb2,int nb3,T* array3D) {

}

template <class T>
void Utils<T>::getArrayDeviceToHost(int nbBloc,int nbThread,int size,T * inDeviceArray, T *outHostArray,cudaStream_t * stream) {
	cudaError_t err ;
	if ( stream != NULL ) {
		err = cudaMemcpyAsync(outHostArray,inDeviceArray, size*sizeof(T), cudaMemcpyDeviceToHost,*stream);
	} else {
		err = cudaMemcpy(outHostArray,inDeviceArray, size*sizeof(T), cudaMemcpyDeviceToHost);
	}
	safecall(err);
	if ( err ) {
		printf("Error QTLMapStructDeviceWorkLinear::getArrayDeviceToHost = %d \n",err);
		exit(1);
	}
}


/**********************************************************************************************************************************************
 * NFIX  : nombre de niveau fixe a la position (moyenne general, effet polygenique, effets de nuisances)
 * NVAR  : nombre de niveaux varaible a la position (effet qtl ou haplotype)
 * NTNIV : NFIX + NVAR 
 * NPOS  : nombre de position teste 
 * X     : matrice de contingence (animal par ligne / niveau exprime en colonne)
 * XT    : transpose de X
 * XT.X => Calcul independant de block A ( NFIX * NFIX operation ), B ( NFIX * NVAR * NPOS) , C (NVAR * NVAR/2 * NPOS)
 * 
 * Exemple : NFIX = 4, NVAR = 3   
 * 
 *              LEV FIX     POS1        POS2        POS3          POS4
 *           ( 1 1 0 0  0.5 0  0    0.4  0  0    0.3  0  0    0.2  0  0  )      NPOS = 4
 * TOTAL_X = ( 1 1 0 0  0.5 0  0    0.4  0  0    0.3  0  0    0.2  0  0  )
 *           ( 1 0 1 0  0  0.5 0     0  0.4  0    0  0.4  0    0  0.4  0 )    
 *           ( 1 0 1 0  0  0.5 0     0  0.4  0    0  0.4  0    0  0.4  0 )
 *           ( 1 0 0 1  0   0 0.5    0  0  0.4    0  0   0.4   0  0  0.4 )
 *           ( 1 0 0 1  0   0 0.5    0  0  0.4    0  0   0.4   0  0  0.4 )
 * 
 * pour POS = 1
 * 
 *     ( 1 1 0 0 0.5 0  0  )
 * X = ( 1 1 0 0 0.5 0  0  )
 *     ( 1 0 1 0  0 0.5 0  )     ND=6
 *     ( 1 0 1 0  0 0.5 0  )
 *     ( 1 0 0 1  0  0 0.5 )
 *     ( 1 0 0 1  0  0 0.5 )
 * 
 * 
 * XT.X = ( 1.1 1.2 1.3 1.4 1.5 1.6 1.7 )
 *        (     2.2 2.3 2.4 2.5 2.6 2.7 )
 *        (         3.3 3.4 3.5 3.6 3.7 )            ( 1.1 1.2 1.3 1.4 )         ( 1.5 1.6 1.7 )        ( 5.5 5.6 5.7 ) 
 *        (             4.4 4.5 4.6 4.7 )   avec A = (     2.2 2.3 2.4 )    B  = ( 2.5 2.6 2.7 )    C = (     6.6 6.7 )
 *        (                 5.5 5.6 5.7 )            (         3.3 3.4 )         ( 3.5 3.6 3.7 )        (         7.7 )
 *        (                     6.6 6.7 )            (             4.4 )         ( 4.5 4.6 4.7 )        
 *        (                         7.7 )
 * 
 * 
 *  - Calcul de A sur le HOST
 *  - Calcul de B sur le DEVICE
 *  - Calcul de C sur le DEVICE
 *  
 *  
 *  Calcul B :
 *  BLOCK à 2 dimensions : ivar => numero de la colonne dans la liste des niveaux variables
 *                         ifix => numero de la colonne dans la liste des niveaux fixes
 *                         
 *  
 *  Ordonnancement des resultats :
 *  [ifix] => |  1.5 pos1  | 1.6 pos1 |  1.7 pos1 |  1.5 pos2 |   | 1.6 pos2 |  1.7 pos2 ....
 *  
 */

__global__ void XT_X_B(bool hetero,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {

	extern __shared__ DT FIXVALUE[];

	int ivar = blockIdx.x * blockDim.x + threadIdx.x ;
	//int ifix = blockIdx.y ;
	//int isim=1 ; /* pour l instant on gere que pour 1 simul */

	if ( blockIdx.y < constNbLevelFix ) {
		int i_shared = threadIdx.x;
		while(i_shared < data.nd)
		{
			FIXVALUE[i_shared] = data.contingence_fix[blockIdx.y*data.ndmax+i_shared];
			i_shared += blockDim.x;
		}
		__syncthreads();
	}


	if ( ivar < constNbLevelVar*data.npos ) {
		DT v = 0;
		if (hetero) {
			for (int kd=0;kd<data.nd;kd++ ) {
				v += FIXVALUE[kd]*data.contingence_var[kd*constNbLevelVar*data.npos+ivar]*(data.CD[kd] / data.isigsq[data.corIpKd[kd]*data.nsim]) ; //isim + data.corIpKd[kd]*data.nsim
			}
		} else {
			//////#pragma unroll 4
			for (int kd=0;kd<data.nd;kd++ ) {
				v += FIXVALUE[kd]*data.contingence_var[kd*constNbLevelVar*data.npos+ivar];
			}
		}
		//	work.B_res[blockIdx.y*constNbLevelVar*data.npos+ivar] =  v;
		/* niveau à une position */
		int iniv = ivar % constNbLevelVar ;
		/* position teste */
		int ipos = ivar / constNbLevelVar ;
		work.XX[ipos + constCorrIndexCol[constNbLevelFix + iniv]*data.npos +constCorrIndexCol[blockIdx.y]*data.ntniv*data.npos ] = v;
		work.XX[ipos + constCorrIndexCol[blockIdx.y]*data.npos +constCorrIndexCol[constNbLevelFix + iniv]*data.ntniv*data.npos ] = v;
	}
}
/* calcul d'une valeur de la diagonale et des valeurs a droite de cette diagonale de XT.X   = colonne et ligne de niveau variable =  */
/*
 * 
 *     ( 5.5 5.6 5.7 )
 * C = (     6.6 6.7 )
 *     (         7.7 )
 *     
 *     
 *     stockage du resultat dans C_res sous la forme
 *     
 *            POS 1              POS2                   POS N
 *     | 5.5 | 6.6 | 7.7 | 5.5 | 6.6 | 7.7 | ... | 5.5 | 6.6 | 7.7 |
 *     puis
 *            POS 1     POS 2            POS N
 *     | 5.6 | 6.7 | X | 5.6 | 6.7 | X | ... | 5.6 | 6.7 | X |
 *     puis
 *       POS1              POS2               POS N
 *     [ 5.7 | X | X | 5.7 | X | X |.... | 5.7 | X | X |  
 * 
 */
__global__ void XT_X_C(bool hetero,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	int ivar = blockIdx.x * blockDim.x + threadIdx.x ;

	/* ivar : la diagonale de XT . X */

	if ( ivar < constNbLevelVar*data.npos ) {	           
		/* iniv position entre 0 et ntlevVar */ 
		int nfix = constNbLevelFix ;
		int nvar = constNbLevelVar ;
		int ntniv = data.ntniv ;

		/* niveau à une position */
		int iniv = ivar % nvar ;
		/* position teste */
		int ipos = ivar / nvar ;

		/* initialisation */
		DT viniv[MAXND_WORK] ;
		DT v=0.0;

		int isim=1; /* a adapte pour plusieurs simul ...*/

		if (hetero) {
			for (int kd=0;kd<data.nd;kd++ ) {
				viniv[kd] = data.contingence_var[kd*nvar*data.npos+ivar]*(data.CD[kd] / data.isigsq[isim+data.corIpKd[kd]*data.nsim]);
				v += viniv[kd]*viniv[kd];		                                 
			}		
		} else {
			for (int kd=0;kd<data.nd;kd++ ) {
				viniv[kd] = data.contingence_var[kd*nvar*data.npos+ivar];
				v += viniv[kd]*viniv[kd];		                                 
			}		
		}

		/* calcul de la diagonal */

		/* les diagonales sont stockees en premieres...*/ 
		//work.C_res[ivar] = v; /* il y a NVAR*NPOS valeurs stocke en debut de ce tableau C_res....*/
		work.XX[ipos + constCorrIndexCol[nfix + iniv]*data.npos +constCorrIndexCol[nfix + iniv]*ntniv*data.npos ] = v;

		//int scale = 0.0;

		for (int jniv=1;jniv<(nvar-iniv);jniv++) {
			//scale +=  nvar*data.npos;
			DT v = 0.0;	
			if (hetero) {
				for (int kd=0;kd<data.nd;kd++ ) {
					v +=  viniv[kd]*data.contingence_var[kd*nvar*data.npos+(ivar+jniv)]*(data.CD[kd] / data.isigsq[isim+data.corIpKd[kd]*data.nsim]);
				}
			} else {
				for (int kd=0;kd<data.nd;kd++ ) {
					v +=  viniv[kd]*data.contingence_var[kd*nvar*data.npos+(ivar+jniv)];
				}
			}

			//work.C_res[scale+ivar] = v;
			work.XX[ipos + constCorrIndexCol[nfix + jniv + iniv]*data.npos +constCorrIndexCol[nfix + iniv]*ntniv*data.npos ] = v;
			work.XX[ipos + constCorrIndexCol[nfix + iniv]*data.npos +constCorrIndexCol[nfix + jniv + iniv]*ntniv*data.npos ] = v;
		}
	}
}

/*******************************************************************************************************************************************/
/*#define NB_KD_COL 4 
 * #define MAXNBVAR_WORK 50
__global__ void XT_X_C_VERS2(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) // FLOAT* mat, FLOAT* res
{
	__shared__ DT vect[BLOCKDIMX_XTX_C][NB_KD_COL];
   DT part_res[BLOCKDIMX_XTX_C][MAXNBVAR_WORK/BLOCKDIMX_XTX_C+1];

   int nfix = constNbLevelFix ;
   int nvar = constNbLevelVar ;

   for(int load = 0;load < nvar; load += BLOCKDIMX_XTX_C)
   {
       bool last_iteration = false;
       for(int l=0; l < BLOCKDIMX_XTX_C; l++)
       {
           #pragma unroll 4
           for(int k=0; k< (nvar/BLOCKDIMX_XTX_C+1); k++)
           {
               part_res[l][k] = 0;
           }
       }
       for(int shared_load =0; shared_load < data.nd; shared_load+=NB_KD_COL)
       {
           int shared_loaded = min(NB_KD_COL, data.nd-shared_load);
           if((shared_load+shared_loaded)==data.nd)
           {
               last_iteration = true;
           }
           #pragma unroll 4
           for(int i=0; i < shared_loaded; i++)
           {
               if(load + threadIdx.x < nvar)
                vect[threadIdx.x][i] = data.contingence_var[(shared_load+i)*__umul24(nvar,data.npos) + __umul24(blockIdx.x,nvar) + load + threadIdx.x];
           }

           __syncthreads();

           int loaded = min(BLOCKDIMX_XTX_C, nvar-load);
           for(int var = 0; var < loaded;var++)
           {
               for(int var2 = 0, ite = 0; var2+threadIdx.x < nvar; var2+=BLOCKDIMX_XTX_C, ite++)
               {
                   #pragma unroll 4
                   for(int j=0;j < shared_loaded;j++)
                   {
                       part_res[var][ite] += vect[var][j]*data.contingence_var[(shared_load+j)*__umul24(nvar,data.npos) + __umul24(blockIdx.x,nvar) +var2+threadIdx.x];
                   }
                   if(last_iteration)
                   {
                	  int iniv = var2 + threadIdx.x ;
                	  int jniv = var+load ;
                	  int ipos = blockIdx.x ;

                      //res[(var+load)*__umul24(nvar,data.npos) + __umul24(blockIdx.x,nvar) + var2 + threadIdx.x] = 
                      // work.XX[ipos + data.corrIndexCol[nfix + jniv + iniv]*data.npos +data.corrIndexCol[nfix + iniv]*(nfix+nvar)*data.npos ] = part_res[var][ite];
                      //  work.XX[ipos + data.corrIndexCol[nfix + iniv]*data.npos +data.corrIndexCol[nfix + jniv + iniv]*(nfix+nvar)*data.npos ] = part_res[var][ite];   			
                   }
               }
           }
       }
   }
} 
 */
/**
 *
 * Cholesky_Decomposition
 *
 *
 *
 */
__global__ void Cholesky_Decomposition(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {

	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */

	DT * triang = work.triang ;
	int ntniv = data.ntniv;
	int nfix = constNbLevelFix;

	if ( ipos < data.npos ) {

		for (int i=0;i<ntniv;i++) {
			for (int j=0;j<ntniv;j++) {
				triang[ipos+i*data.npos+j*ntniv*data.npos] = 0.0;
			}
		}
		/* a optimiser , ces valeurs sont redondantes dans les matrices XX pour chaque position  */
		for (int i=0;i<nfix;i++) {
			for (int j=i;j<nfix;j++) {
				DT v = work.A_res[j*nfix+i];
				work.XX[ipos+constCorrIndexCol[j]*data.npos +constCorrIndexCol[i]*ntniv*data.npos] = v;
				work.XX[ipos+constCorrIndexCol[i]*data.npos +constCorrIndexCol[j]*ntniv*data.npos] = v;
			}
		}

		for (int j=0;j<ntniv;j++) {
			DT v = work.XX[ipos+j*data.npos +j*ntniv*data.npos]; //XX[j][j] ;

			for (int k=0;k<=j-1;k++) {
				// [j][j] = [j][j] - [k][j]*[k][j]
				v -= triang[ipos+k*data.npos+j*ntniv*data.npos]*triang[ipos+k*data.npos+j*ntniv*data.npos];
			}
			//si estimable .. on ne met pas de structure vecseuil==> on utilise la diagonale de triangle pour savoir l effet est estimable
			triang[ipos+j*data.npos +j*ntniv*data.npos] =  sqrt(v);
			if ( triang[ipos+j*data.npos +j*ntniv*data.npos] != triang[ipos+j*data.npos +j*ntniv*data.npos] ) triang[ipos+j*data.npos +j*ntniv*data.npos] = 0.0;

			if ( triang[ipos+j*data.npos +j*ntniv*data.npos] > data.seuil_cho ) { 
				for (int i=j+1;i<ntniv;i++ ) {
					triang[ipos+i*data.npos +j*ntniv*data.npos] = work.XX[ipos+i*data.npos +j*ntniv*data.npos];//XX[i][j];
					for ( int k=0;k<=j-1;k++) {
						triang[ipos+i*data.npos +j*ntniv*data.npos] = triang[ipos+i*data.npos +j*ntniv*data.npos] -  (triang[ipos+i*data.npos +k*ntniv*data.npos] * triang[ipos+j*data.npos +k*ntniv*data.npos]);
					}
					triang[ipos+i*data.npos +j*ntniv*data.npos] = triang[ipos+i*data.npos +j*ntniv*data.npos] / triang[ipos+j*data.npos +j*ntniv*data.npos];
					triang[ipos+j*data.npos +i*ntniv*data.npos] = triang[ipos+i*data.npos +j*ntniv*data.npos] ;
				}
			}  
		}
	}

}

/*
 * A optimiser : rhs pour les effet fixe a la position ne devrait prendre que nsim*nlevelfix et non pas nsim*nlevelfix*npos
 * 
 * 
 * logique de corrIndexCol :
 * 
 *     I=1,..,NTNIV | I=1,..,NFIX,NFIX+1,..,NVAR
 *     -------------------------------------------
 *     0            |            0
 *     1            |            J
 *     2            |            J+1
 *     3            |            .
 *     .            |            1
 *     .            |            2
 *     .            |            3
 *     NTNIV        |            NTNIV
 * 
 * 
 * 
 */
__global__ void Set_RHS_FIX(bool hetero,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	int isim = blockIdx.y * blockDim.y + threadIdx.y ;// * blockDim.y + threadIdx.y ;

	int ntniv = data.ntniv ;
	int idNiv;
	if ( ipos < data.npos && isim < data.nsim ) {
		for (int iniv=0;iniv<constNbLevelFix;iniv++) {

			idNiv = constCorrIndexCol[iniv];
			
			int index = ipos + isim*data.npos + idNiv*data.nsim*data.npos ;
			work.rhs[index]=0;
			if (work.triang[ipos+idNiv*data.npos+idNiv*data.npos*ntniv]>data.seuil_cho) {
				if ( hetero ) {
					for (int kd = 0;kd<data.nd;kd++ ) {
						work.rhs[index] += data.contingence_fix[iniv*data.ndmax + kd]*data.Y[isim+kd*data.nsim]*(data.CD[kd] / data.isigsq[isim+data.corIpKd[kd]*data.nsim]);
					}
				} else {
					for (int kd = 0;kd<data.nd;kd++ ) {
						work.rhs[index] += data.contingence_fix[iniv*data.ndmax + kd]*data.Y[isim+kd*data.nsim];
					}
				}
			}
		}
	}
}

__global__ void Set_RHS_VAR(bool hetero,QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	int ivar = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	int isim = blockIdx.y ;// * blockDim.y + threadIdx.y ;

	if ( ivar < constNbLevelVar*data.npos ) {	
		/* iniv position entre 0 et ntlevVar */ 
		int ntniv = data.ntniv ;

		/* niveau à une position */
		int iniv = ivar % constNbLevelVar ;
		/* position teste */
		int ipos = ivar / constNbLevelVar ;

		//for (int isim=0;isim<data.nsim;isim++) {
		int index = ipos + isim*data.npos + constCorrIndexCol[constNbLevelFix+iniv]*data.nsim*data.npos ;
		work.rhs[index]=0;
		if (work.triang[ipos+constCorrIndexCol[constNbLevelFix+iniv]*data.npos+constCorrIndexCol[constNbLevelFix+iniv]*data.npos*ntniv]>data.seuil_cho) {
			if (hetero) {
				for (int kd = 0;kd<data.nd;kd++ ) {
					work.rhs[index] += data.contingence_var[kd*constNbLevelVar*data.npos+ivar]*data.Y[isim+kd*data.nsim]*(data.CD[kd] / data.isigsq[isim+data.corIpKd[kd]*data.nsim]);
				}
			} else {
				for (int kd = 0;kd<data.nd;kd++ ) {
					work.rhs[index] += data.contingence_var[kd*constNbLevelVar*data.npos+ivar]*data.Y[isim+kd*data.nsim];
				}
			}
		}
		//}
	}
}

/**
 *
 * Resolution LU (Partie descendante descendante )
 *
 *
 */
__global__ void Resolve_LU(QTLMapStructDeviceDataLinear data, 
		QTLMapStructDeviceWorkLinear work,
		QTLMapStructDeviceSolutionLinear solution) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	int isim = blockIdx.y * blockDim.y + threadIdx.y ;

	int ntniv = data.ntniv ;
	DT temp[MAXNTNIV_LOCALTHREAD_MEMORY];

	/******* DESCENT LU *****/
	if ( ipos < data.npos && isim < data.nsim ) {
		//for (int isim=0;isim<data.nsim;isim++) {
		int scale = ipos + isim*data.npos ;
		for (int i=0;i<ntniv;i++) {
			temp[i] = 0.0;
			if (work.triang[ipos+i*data.npos+i*ntniv*data.npos]>data.seuil_cho) {
				temp[i] = work.rhs[scale + i*data.nsim*data.npos];
				for (int j=i-1;j>=0;j--) {
					if (work.triang[ipos+j*data.npos+j*ntniv*data.npos]>data.seuil_cho) {
						temp[i] = temp[i] - temp[j]*work.triang[ipos+j*data.npos+i*ntniv*data.npos];
					}
				}
				temp[i] = temp[i] / work.triang[ipos+i*data.npos+i*ntniv*data.npos];
			} 
		}

		for (int i=ntniv-1;i>=0;i--) {
			int indexI = scale + i*data.nsim*data.npos;
			solution.bestim[indexI] = 0.0 ;
			if (work.triang[ipos+i*data.npos+i*ntniv*data.npos]>data.seuil_cho) {
				solution.bestim[indexI] = temp[i];
				for (int j=i+1;j<ntniv;j++) {
					if (work.triang[ipos+j*data.npos+j*ntniv*data.npos]>data.seuil_cho) {
						int indexJ = scale + j*data.nsim*data.npos;
						solution.bestim[indexI] -= solution.bestim[indexJ]*work.triang[ipos+i*data.npos+j*ntniv*data.npos];
					}
				}
				solution.bestim[indexI] = solution.bestim[indexI] / work.triang[ipos+i*data.npos+i*ntniv*data.npos]; 
			} 
		}
		//}//fin for isim

	}
}

/**
 *
 * Set_SIGSQ
 * XB = Y - ( X' . Bestim )
 * SIGSQ = SUM(XB^2)
 */
__global__ void Set_SIGSQ(QTLMapStructDeviceDataLinear data,
		QTLMapStructDeviceWorkLinear work,
		QTLMapStructDeviceSolutionLinear solution) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	int isim = blockIdx.y * blockDim.y + threadIdx.y ; /* simulation */

	int ntniv = data.ntniv ;

	if ( ipos < data.npos && isim < data.nsim ) {
		//	for (int isim=0;isim<data.nsim;isim++) {
		int scale = ipos + isim*data.npos ;
		DT xb = 0 ;

		for (int kd = 0;kd<data.nd;kd++ ) {
			DT v=0;

#pragma unroll 4
			for (int iniv=0;iniv<constNbLevelFix;iniv++) {
				v += data.contingence_fix[iniv*data.ndmax+kd]*solution.bestim[scale + constCorrIndexCol[iniv]*data.nsim*data.npos];
			}

			for (int iniv=constNbLevelFix;iniv<ntniv;iniv++) {	
				v += data.contingence_var[kd*constNbLevelVar*data.npos+ipos*constNbLevelVar+(iniv-constNbLevelFix)]*solution.bestim[scale + constCorrIndexCol[iniv]*data.nsim*data.npos];
			}

			v = data.Y[isim+kd*data.nsim] - v;
			xb += v*v ;
		}
		DT v = xb / data.nd ;

#pragma unroll 4
		for (int ip=0;ip<data.np;ip++) {
			solution.osigsq[ipos+isim*data.npos+ip*data.npos*data.nsim] = v ;
		}
	}
	//}
}


/**
 *
 * Set_SIGSQ
 * XB = Y - ( X' . Bestim )
 * SIGSQ = SUM(XB^2)
 */
__global__ void Set_SIGSQ_HETERO(QTLMapStructDeviceDataLinear data,
		QTLMapStructDeviceWorkLinear work,
		QTLMapStructDeviceSolutionLinear solution) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */

	int ntniv = data.ntniv ;

	if ( ipos < data.npos ) {
		for (int isim=0;isim<data.nsim;isim++) {
			int scale = ipos + isim*data.npos ;
			DT xb[MAXNP_WORK] ;
			for (int ip=0;ip<data.np;ip++) {
				xb[ip]=0;
			}

			for (int kd = 0;kd<data.nd;kd++ ) {
				DT v=0;

				for (int iniv=0;iniv<constNbLevelFix;iniv++) {
					v += data.contingence_fix[iniv*data.ndmax+kd]*solution.bestim[scale + constCorrIndexCol[iniv]*data.nsim*data.npos];
				}

				for (int iniv=constNbLevelFix;iniv<ntniv;iniv++) {	
					v += data.contingence_var[kd*constNbLevelVar*data.npos+ipos*constNbLevelVar+(iniv-constNbLevelFix)]*solution.bestim[scale + constCorrIndexCol[iniv]*data.nsim*data.npos];
				}

				v = data.Y[isim+kd*data.nsim] - v;
				//xb += v*v ;
				xb[data.corIpKd[kd]] += v*v*data.CD[kd]*data.CD[kd] ;
			}

			for (int ip=0;ip<data.np;ip++) {
				solution.osigsq[ipos+isim*data.npos+ip*data.npos*data.nsim] = xb[ip] / constSizeFam[ip];
			}
		}
	}
}

/**
 *
 * Set_LRT
 *
 * 
 */
__global__ void Set_LRT(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,QTLMapStructDeviceSolutionLinear solution,int scalePosition) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */

	if ( ipos < data.npos ) {
		for (int isim=0;isim<data.nsim;isim++) {
			solution.lrt[scalePosition+ipos+isim*data.nposGlobal] = 0 ;
			int k = ipos + isim * data.npos ;
			DT v = 0;
			for (int ip=0;ip<data.np;ip++) {
				v += constSizeFam[ip] * (log(data.isigsq[isim+ip*data.nsim]) - log(solution.osigsq[k+ip*data.npos*data.nsim]));
				//v=solution.osigsq[k+ip*data.npos*data.nsim];
			}
			solution.lrt[scalePosition+ipos+isim*data.nposGlobal] = v;
		}//fin for isim
	}
}



/*
  xinc  = matrice d'incidence
  nkd   = nombre de ligne
  ntniv = nombre de colonne
  seuil_choPtr = seuil d'estimabilite pour la solution

  column major order ===> offset = row + column*NUMROWS
  XINC ACCESS => ndmax*npos*iniv+npos*kd+ipos

  XINC(construction : level fixe puis levels variables)
  Mu MuPere1 MuPere2 MuMere1 MuMere2 QTLPere1Pos1 QTLPere2Pos1 QTLMere1Pos1 QTLMere2Pos1 ... QTLPere1PosN QTLPere2PosN QTLMere1PosN QTLMere2PosN

  nLevelFixPtr : 5 (Moyenne gen + 4 effet polygenique)
  nLevelVarPtr : 4 (4 effet qtl a estimer)
  corrLevelColPtr(1) = 1    moyenne generale
  corrLevelColPtr(2) = 6    effet qtl pere 1
  corrLevelColPtr(3) = 7    effet qtl pere 2
  corrLevelColPtr(4) = 8    effet qtl mere 1
  corrLevelColPtr(5) = 9    effet qtl mere 2
  corrLevelColPtr(6) = 2    effet polygenique pere 1
  corrLevelColPtr(7) = 3    effet polygenique pere 2
  corrLevelColPtr(8) = 4    effet polygenique mere 1
  corrLevelColPtr(9) = 5    effet polygenique mere 2
 */
extern "C" void cuda_homoscedatic_resolv_genome_(
		int *nqtlPtr,                         /* IN  : Number of QTL */
		DT  *sigsquare,                       /* IN  : variance residuelle sous les hypothese 0 à NQTL-1 tableau [NSIM][NP] */
		int *nLevelFixPtr,                    /* IN  : nombre de niveaux fixe a la position : moyenne genrale, polygenic, effet fixe, covariables */
		int *nLevelVarPtr,                    /* IN  : nombre de niveau variable a la position */
		int *corrLevelColPtr,                 /* IN  : Correspondance du niveau associe a une colonne de la matrice de contingence */
		void *work_cuda,                      /* IN  : structure de travail pour la fonction get_partialXinc */
		void (*get_partialXincFix)(void*,DT*),   /* IN  : Fonction qui donne les donnees fixe a la position de la matrice d incidences... */
		void (*get_partialXincVar)(void*,int*,int*,DT*),            /* IN  : Fonction qui donne les donnees variables a la position des matrices d incidences... */
		DT *Y_d,                              /* IN  : performance [nsim,nd] */
		DT *CD_d,                             /* IN  : cd of anim  [nd]      */
		int *nsimPtr,                         /* IN  : nombre d'individu total  : nombre de ligne reelle de XINC */
		int *ndPtr,                           /* IN  : nombre d'individu total  : nombre de ligne reelle de XINC */
		int *nkdPtr,                          /* IN  : nombre d'individu pris en compte dans les matrices de contingence */
		int *npPtr,                           /* IN  : nombre de famille de pere */
		int *sizeFamilyNp,                    /* IN  : nombre de descendant par pere tableau [NP] */
		int *ntnivmaxPtr,                     /* IN  : nombre maximum de niveau : nombre de colonne reelle de XINC */
		int *npositionPtr,                    /* IN  : number of position estimation */
		double  *seuil_choPtr,                /* IN  : seuil pour considere grace a la decomposition de cholesky si un niveau est estimable */
		//		int    * vecsol,                      /* OUT : tableau de booleen de taille [NPOSITION,NTNIV]  pour l'estimabilite de chaque niveau */
		DT * bestim,                          /* OUT : tableau solution de l'estimation de chaque niveau (taille [NPOSITION,NSIM,NTNIV]) */
		DT * osigsquare,                      /* OUT : tableau des variances residuelles pour chaque familles de pere (taille [NPOSITION,NSIM,NP])  */
		DT * lrt,                             /* OUT : LRT (taille [NPOSITION,NSIM])  */
		DT * maxLRT,                          /* OUT : Value of the Maximul LRT  [NSIM] */
		int *maxPosition                      /* OUT : Index of the position maximum tableau [NSIM] */
) {
	bool hetero = false;

	QTLMapStructDeviceDataLinear      data ;
	QTLMapStructDeviceWorkLinear      work ;
	QTLMapStructDeviceSolutionLinear  solution ;
	
	safecall(cudaMemcpyToSymbol(constNbLevelFix, nLevelFixPtr, sizeof(int)));
	safecall(cudaMemcpyToSymbol(constNbLevelVar, nLevelVarPtr, sizeof(int)));
	safecall(cudaMemcpyToSymbol(constNSim, nsimPtr, sizeof(int)));
	
	if ( ( (*nLevelFixPtr) + (*nLevelVarPtr)) >= MAXNTNIV_LOCALTHREAD_MEMORY ) {
		cerr << "Can not initialised cuda structure => update constant MAXNTNIV_LOCALTHREAD_MEMORY :"<<  MAXNTNIV_LOCALTHREAD_MEMORY << "<=" <<  (*nLevelFixPtr + *nLevelVarPtr) << endl ;
		exit(1);
	}

	safecall(cudaMemcpyToSymbol(constCorrIndexCol, corrLevelColPtr, (*nLevelFixPtr + *nLevelVarPtr)*sizeof(DT)));
	
	if ( (*npPtr) >=  MAXNP_WORK ) {
			cerr << "Can not initialised cuda structure => update constant MAXNP_WORK :"<<  MAXNP_WORK << "<=" <<  (*npPtr) << endl ;
			exit(1);
		}
	
	cudaMemcpyToSymbol(constSizeFam,sizeFamilyNp,(*npPtr)*sizeof(int));
	
#if(_HOST_DEBUG_)
	printf("****************** INFO DEVICE ******************************\n");
	int nbDevice ;
	cudaGetDeviceCount(&nbDevice);
	printf("NB DEVICE     : %d\n",nbDevice);

	cudaDeviceProp  prop;

	for (int i=0;i<nbDevice;i++) {
		cudaGetDeviceProperties(&prop,i);
		printf(" ******* ===  %s  ===  *******\n",prop.name);
		printf(" ID DEVICE:%d\n",i);
		printf("  totalGlobalMem : %lu \n",prop.totalGlobalMem);
		printf("  sharedMemPerBlock : %lu \n",prop.sharedMemPerBlock);
		printf("  regsPerBlock : %d \n",prop.regsPerBlock);
		printf("  warpSize : %d \n",prop.warpSize);
		printf("  maxThreadsPerBlock : %d \n",prop.maxThreadsPerBlock);
		printf("  maxThreadsDim : %d,%d,%d \n",prop.maxThreadsDim[0],prop.maxThreadsDim[1],prop.maxThreadsDim[2]);
		printf("  maxGridSize : %d,%d,%d \n",prop.maxGridSize[0],prop.maxGridSize[1],prop.maxGridSize[2]);
		printf("  clockRate : %d \n",prop.clockRate);
		printf("  totalConstMem : %lu \n",prop.totalConstMem);
		printf("  multiProcessorCount : %d \n",prop.multiProcessorCount);
	}
	print_info_memory();
#endif

	int err ;

	DT *xinc_d_var,*xinc_d_fix;

	/* Initialisation des donnees fixes a la position de la matrice de contingence */
	safecall(cudaMallocHost(& xinc_d_fix, (*ndPtr)*((*nLevelFixPtr)*sizeof(DT))));
	get_partialXincFix(work_cuda,xinc_d_fix);

	data.init(nqtlPtr,sigsquare,xinc_d_fix,Y_d,CD_d,corrLevelColPtr,seuil_choPtr,ndPtr,nkdPtr,nsimPtr,npositionPtr,nLevelFixPtr,nLevelVarPtr,npPtr,sizeFamilyNp);

	int nbPosBlock;
	nbPosBlock = data.calculBlockPositionWorkSize();
	if ( nbPosBlock > *npositionPtr ) {
		nbPosBlock = *npositionPtr ;
	} 
	data.npos = nbPosBlock ;
	work.initResolution(hetero,xinc_d_fix,data,CD_d,sigsquare,sizeFamilyNp) ;
	solution.init(data);

	/*
	 * 
	 *    Boucle des traitements par block
	 *    Utilisation de stream cuda
	 * 
	 * 
	 */
	/* Variables de travail pour recolter les resultats intermediaires */
	DT *local_osigsquare;
	DT *local_bestim;//[nbPosBlock*data.nsim*data.ntniv];

	safecall(cudaMallocHost(& local_osigsquare, nbPosBlock*data.nsim*data.np*sizeof(DT)));
	safecall(cudaMallocHost(& local_bestim, nbPosBlock*data.nsim*data.ntniv*sizeof(DT)));

	/* Creation des streams */
	int nstream=2;
	cudaStream_t stream[nstream];
	for (int i=0;i<nstream;i++) {
		cudaStreamCreate(&stream[i]);
	}
	int istream = 0 ;
	/* premier appel */

	int sizeBlockXincDVar = 0; /* taille en nombre de position a teste sur le groupe de liaison */
	int nbBlockXincDVar=2; /* On gere 2 block .l'un a remplir, l autre pour le calcul */
	int currentBlockMod = 0; 
	if ((*nLevelVarPtr) > 0) {
		/* Initialisation des donnees variables a la position de la matrice de contingence */
		sizeBlockXincDVar = (*ndPtr)*((*nLevelVarPtr)*nbPosBlock) ;
		safecall(cudaMallocHost(& xinc_d_var, nbBlockXincDVar*sizeBlockXincDVar*sizeof(DT)));
	}

	int lastPosition = 1;
	int nextPosition = nbPosBlock;
	get_partialXincVar(work_cuda,&lastPosition,&nextPosition,xinc_d_var);
	print_info_memory();

/*
 * Initialisation pour benchmark
 */	
#if(_HOST_TIME_PROF_)
	cudaEvent_t start, stop;
	float elapsedTime;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
#endif

	for (int lastPosition = 1;lastPosition<=data.nposGlobal;lastPosition += nbPosBlock){   
#pragma omp sections
		{
#pragma omp section
			{
				int lastPositionNext = lastPosition + nbPosBlock ;
				if ( lastPositionNext <=data.nposGlobal ) {
					int nextPosition = lastPositionNext + nbPosBlock - 1;
					if ( nextPosition > data.nposGlobal) {
						nextPosition = data.nposGlobal;
					}

					int nextBlockMod = (currentBlockMod+1)%nbBlockXincDVar;
					get_partialXincVar(work_cuda,&lastPositionNext,&nextPosition,xinc_d_var+(nextBlockMod)*sizeBlockXincDVar);
				}
			}
#pragma omp section
			{
				nextPosition = lastPosition + nbPosBlock - 1;
				if ( nextPosition > data.nposGlobal) {
					nextPosition = data.nposGlobal;
				}
				data.npos = nextPosition - lastPosition + 1 ;

				cout << "   ** compute from position:"<< lastPosition << " at:"<< nextPosition << " (NPOS:"<< data.nposGlobal << ")   **" << endl << endl;

				/* Type de block utilise par les kernels */
				int nbblock_512 = ceil(data.npos / MAX_BLOCKDIM_512) + 1 ;
				int nbblock_64  = ceil(data.npos / MAX_BLOCKDIM_64) + 1 ;

				data.set_contingence_var(*ndPtr,*nLevelVarPtr,data.npos,0,xinc_d_var+currentBlockMod*sizeBlockXincDVar,&stream[istream]);//

				if ( data.ntniv > MAXNTNIV_LOCALTHREAD_MEMORY ) {
					printf("***************************>>>>ntniv too big......maximum allowed :%d",MAXNTNIV_LOCALTHREAD_MEMORY);
					exit(1);
				}

				err = (int)cuInit(0);
				if ( err ) {
					printf(" cuInit error : %d \n",err);
					exit(1);
				}

#if(_HOST_TIME_PROF_)
	cudaEventRecord(start, 0);
#endif
				// ***********************************
				// Allocate vectors in device memory
				// ***********************************

				dim3 dimBlock(BLOCKDIMX);
				dim3 dimGrid (ceil(data.npos*data.nLevelVar / dimBlock.x) + 1 , (data.nLevelFix / dimBlock.y) + 1);

				XT_X_B<<<dimGrid,dimBlock,data.nd*sizeof(DT),stream[istream]>>>(hetero,data,work);
				cudaStreamSynchronize(stream[istream]);
#if(_HOST_TIME_PROF_)
	cudaEventRecord(stop, 0);
	cudaEventElapsedTime(&elapsedTime, start, stop);
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
	cout << "[BENCHMARK]  XT_X_B grid(x,y,z):"<< dimGrid.x << "," << dimGrid.y << "," << dimGrid.z << " block(x,y,z):" << 
			dimBlock.x << "," << dimBlock.y << "," << dimBlock.z << " elapsedtime:"<< elapsedTime<< endl ;
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif
				safecall(cudaGetLastError());

#if(_HOST_TIME_PROF_)
	cudaEventRecord(start, 0);
#endif
				dim3 dimGrid2 (ceil(data.npos*data.nLevelVar / dimBlock.x) + 1 , 1);
				XT_X_C<<<dimGrid2,dimBlock,0,stream[istream]>>>(hetero,data,work);

				cudaStreamSynchronize(stream[istream]);
				safecall(cudaGetLastError());
				
#if(_HOST_TIME_PROF_)
	cudaEventRecord(stop, 0);
	cudaEventElapsedTime(&elapsedTime, start, stop);
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
	cout << "[BENCHMARK]  XT_X_C grid(x,y,z):"<< dimGrid2.x << "," << dimGrid2.y << "," << dimGrid2.z << " block(x,y,z):" << 
			dimBlock.x << "," << dimBlock.y << "," << dimBlock.z << " elapsedtime:"<< elapsedTime<< endl ;
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif
#if(_HOST_DEBUG_)
				printf("X'.X\n----------------------------------------------------------\n");
				Utils<DT>::printFloatDeviceArray3D(data.npos,data.ntniv,data.ntniv,1,data.ntniv,data.ntniv,work.XX);
#endif
				
#if(_HOST_TIME_PROF_)
	cudaEventRecord(start, 0);
#endif
				/* Cholesky decomp */
				Cholesky_Decomposition<<<nbblock_512,MAX_BLOCKDIM_512,0,stream[istream]>>>(data,work);
				cudaStreamSynchronize(stream[istream]);

#if(_HOST_TIME_PROF_)
	cudaEventRecord(stop, 0);
	cudaEventElapsedTime(&elapsedTime, start, stop);
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
	cout << "[BENCHMARK]  Cholesky_Decomposition grid(x,y,z):"<< nbblock_512 << "," << 0 << "," << 0 << " block(x,y,z):" << 
	MAX_BLOCKDIM_512 << "," << 0 << "," << 0 << " elapsedtime:"<< elapsedTime<< endl ;
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif
				safecall(cudaGetLastError());

#if(_HOST_DEBUG_)
				printf("Cholesky decomp\n----------------------------------------------------------\n");
				Utils<DT>::printFloatDeviceArray3D(data.npos,data.ntniv,data.ntniv,1,data.ntniv,data.ntniv,work.triang);
#endif
#if(_HOST_TIME_PROF_)
	cudaEventRecord(start, 0);
#endif
				/* RHS */
//				dim3 dimGrid_Set_RHS (ceil(data.npos*data.nLevelVar / MAX_BLOCKDIM_64) + 1 ,data.nsim);
//				Set_RHS_FIX<<<dimGrid_Set_RHS,MAX_BLOCKDIM_64,0,stream[istream]>>>(hetero,data,work);
								
				dim3 dimBlock_Set_RHS_FIX(16,32);
				dim3 dimGrid_Set_RHS_FIX (ceil(data.npos*data.nLevelVar /dimBlock_Set_RHS_FIX.x ) + 1 ,ceil(data.nsim / dimBlock_Set_RHS_FIX.y )+1);
				Set_RHS_FIX<<<dimGrid_Set_RHS_FIX,dimBlock_Set_RHS_FIX,0,stream[istream]>>>(hetero,data,work);
				
				dim3 dimGrid_Set_RHS_VAR (ceil(data.npos*data.nLevelVar / MAX_BLOCKDIM_64) + 1 ,data.nsim);
				
				Set_RHS_VAR<<<dimGrid_Set_RHS_VAR,MAX_BLOCKDIM_64,0,stream[istream]>>>(hetero,data,work);
				cudaStreamSynchronize(stream[istream]);
				
#if(_HOST_TIME_PROF_)
	cudaEventRecord(stop, 0);
	cudaEventElapsedTime(&elapsedTime, start, stop);
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
	cout << "[BENCHMARK]  Set_RHS_FIX + Set_RHS_VAR grid(x,y,z):"<< " elapsedtime:"<< elapsedTime<< endl ;
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif
				safecall(cudaGetLastError());

#if(_HOST_DEBUG_)
				printf("RHS\n-------POS=1,SIM=nsim => NTNIV---------------------------------------------------\n");
				Utils<DT>::printFloatDeviceArray3D(data.npos,data.nsim,data.ntniv,1,1,data.ntniv,work.rhs);
#endif		

#if(_HOST_TIME_PROF_)
	cudaEventRecord(start, 0);
#endif
				/* Solution */
	            dim3 Resolve_LU_dimBlock(16,32);
				dim3 Resolve_LU_dimGrid (ceil(data.npos/Resolve_LU_dimBlock.x)+1 , ceil(data.nsim / Resolve_LU_dimBlock.y)+1);
                
//                dim3 Resolve_LU_dimGrid (ceil(data.npos/MAX_BLOCKDIM_64)+1 , data.nsim);
//                dim3 Resolve_LU_dimBlock(MAX_BLOCKDIM_64,1);
				Resolve_LU<<<Resolve_LU_dimGrid,Resolve_LU_dimBlock,0,stream[istream]>>>(data,work,solution);
				cudaStreamSynchronize(stream[istream]);

#if(_HOST_TIME_PROF_)
	cudaEventRecord(stop, 0);
	cudaEventElapsedTime(&elapsedTime, start, stop);
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
	cout << "[BENCHMARK]  Resolve_LU grid(x,y,z):"<< Resolve_LU_dimGrid.x << "," << Resolve_LU_dimGrid.y << "," << Resolve_LU_dimGrid.z << " block(x,y,z):" << 
			MAX_BLOCKDIM_64 << "," << 0 << "," << 0 << " elapsedtime:"<< elapsedTime<< endl ;
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif
	
				safecall(cudaGetLastError());

#if(_HOST_DEBUG_)
				printf("Resolution LU\n------POS=1,SIM=nsim => NTNIV ----------------------------------------------------\n");
				Utils<DT>::printFloatDeviceArray3D(data.npos,data.nsim,data.ntniv,1,1,data.ntniv,solution.bestim);
#endif

				/* Recuperation des resultats */
				Utils<DT>::getArrayDeviceToHost(nbblock_512,MAX_BLOCKDIM_512,data.npos*data.nsim*data.ntniv,solution.bestim,local_bestim,&stream[istream]);
				
#if(_HOST_TIME_PROF_)
	cudaEventRecord(start, 0);
#endif
				/* Variances residuelles */
				if (hetero) {
					Set_SIGSQ_HETERO<<<nbblock_64,MAX_BLOCKDIM_64,0,stream[istream]>>>(data,work,solution);
				} else {
//					dim3 dimGrid (ceil(data.npos / MAX_BLOCKDIM_64) + 1 , data.nsim);
//					dim3 dimBlock(MAX_BLOCKDIM_64);
					dim3 dimBlock(16,32);
					dim3 dimGrid (ceil(data.npos/dimBlock.x)+1 , ceil(data.nsim / dimBlock.y)+1);
					
					Set_SIGSQ<<<dimGrid,dimBlock,0,stream[istream]>>>(data,work,solution);
				}
                
				/* on synchronise pour que local_bestim soit initialise */
				cudaStreamSynchronize(stream[istream]);

#if(_HOST_TIME_PROF_)
	cudaEventRecord(stop, 0);
	cudaEventElapsedTime(&elapsedTime, start, stop);
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
	cout << "[BENCHMARK]  Set_SIGSQ grid(x,y,z):"<< nbblock_64 << "," << 0 << "," << 0 << " block(x,y,z):" << 
			MAX_BLOCKDIM_64 << "," << 0 << "," << 0 << " elapsedtime:"<< elapsedTime<< endl ;
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif
	
				for (int ipos=0;ipos<data.npos;ipos++) {
					for (int isim=0;isim<data.nsim;isim++) {
						for (int iniv=0;iniv<data.ntniv;iniv++) {
							bestim[ipos+lastPosition-1+isim*data.nposGlobal+iniv*data.nsim*data.nposGlobal] = local_bestim[ipos+isim*data.npos+iniv*data.nsim*data.npos] ;
						}
					}
				}

				cudaStreamSynchronize(stream[istream]);
				safecall(cudaGetLastError());

#if(_HOST_DEBUG_)
				cout << "SIGSQ------------POS=1,SIM=nsim => NP----------------------------------------------"<<endl ;
				Utils<DT>::printFloatDeviceArray3D(data.npos,data.nsim,data.np,1,data.nsim,data.np,solution.osigsq);
#endif
				/* Recuperation des resultats */
				Utils<DT>::getArrayDeviceToHost(nbblock_512,MAX_BLOCKDIM_512,data.npos*data.nsim*data.np,solution.osigsq,local_osigsquare,&stream[istream]);

				if ( data.nqtl>=1 ) {	
#if(_HOST_TIME_PROF_)
	cudaEventRecord(start, 0);
#endif
					/* Calcul LRT */
					Set_LRT<<<nbblock_512,MAX_BLOCKDIM_512,0,stream[istream]>>>(data,work,solution,lastPosition-1);

#if(_HOST_DEBUG_)
					printf("PARTIEL LRT\n------------POS=1,SIM=1,..,nsim => NP----------------------------------------------\n");
					cudaStreamSynchronize(stream[istream]);
					Utils<DT>::printFloatDeviceArray2D(data.nposGlobal,data.nsim,1,data.nsim,solution.lrt);
#endif
					safecall(cudaGetLastError());
				}

				cudaStreamSynchronize(stream[istream]);

#if(_HOST_TIME_PROF_)
	cudaEventRecord(stop, 0);
	cudaEventElapsedTime(&elapsedTime, start, stop);
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
	cout << "[BENCHMARK]  Set_LRT grid(x,y,z):"<< nbblock_512 << "," << 0 << "," << 0 << " block(x,y,z):" << 
			MAX_BLOCKDIM_512 << "," << 0 << "," << 0 << " elapsedtime:"<< elapsedTime<< endl ;
	cout << "----------------------------------------------------------------------------------------------------------------"<< endl ;
#endif
	
				for (int ipos=0;ipos<data.npos;ipos++) {
					for (int isim=0;isim<data.nsim;isim++) {
						for (int ip=0;ip<data.np;ip++) {
							osigsquare[ipos+lastPosition-1+isim*data.nposGlobal+ip*data.nsim*data.nposGlobal] = local_osigsquare[ipos+isim*data.npos+ip*data.nsim*data.npos];//[ip+isim*data.np+ipos*data.nsim*data.np] ;
						}
					}
				}

				//work.getAllSigDevice(data,nbblock_512,MAX_BLOCKDIM_512,osigsquare) ;
				//work.getAllVecsolDevice(data,nbblock_512,MAX_BLOCKDIM_512,vecsol+(lastPosition-1)*data.ntniv) ;

			}
		}
		istream = (istream+1)%nstream;
		currentBlockMod = (currentBlockMod+1)%nbBlockXincDVar ;

	}
	/* Fin boucle principale  */

	for (int i=0;i<nstream;i++) {
		cudaStreamDestroy(stream[i]);
	}


	work.releaseDeviceResolution();
	safecall(cudaFreeHost(xinc_d_fix));
	if ( data.nqtl>=1 ) safecall(cudaFreeHost(xinc_d_var));
	safecall(cudaFreeHost(local_osigsquare));
	safecall(cudaFreeHost(local_bestim));

	/* FIN si H0 */
	if ( data.nqtl>=1 ) {	
		/* Recuperation des LRTs */
		int nbblock_512_global = ceil(data.nposGlobal / MAX_BLOCKDIM_512) + 1 ;
		Utils<DT>::getArrayDeviceToHost(nbblock_512_global,MAX_BLOCKDIM_512,data.nposGlobal*data.nsim,solution.lrt,lrt,NULL);

		for (int isim=0;isim<data.nsim;isim++) {
			maxLRT[isim]      = -999.9;
			maxPosition[isim] = -1;
			for (int i=0;i<data.nposGlobal;i++) {
				if ( maxLRT[isim] < lrt[i+isim*data.nposGlobal]) {
					maxLRT[isim]      = lrt[i+isim*data.nposGlobal];
					maxPosition[isim] = i+1 ; /* index pour fortran */
				}
			}
		}
	}

	
#if(_HOST_TIME_PROF_)
	cudaEventDestroy(start);
	cudaEventDestroy(stop);
#endif
	
#if(_HOST_DEBUG_)
	cout << "DATA RELEASED ON DEVICE" << endl ;
	data.releaseDevice();
#endif

#if(_HOST_DEBUG_)
	cout << "SOLUTION RELEASED ON DEVICE" << endl ;
	solution.releaseDevice();
#endif


#if(_HOST_DEBUG_)
	cout << " ** END ** " << endl ;
#endif
}
